# -*- coding: UTF-8 -*-
#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
A module of functions and classes for configuring NMPC/MHE problems
"""
import enum
from pyomo.environ import SolverFactory
from pyomo.core.base.var import VarData


class ControlInitOption(enum.Enum):
    FROM_PREVIOUS = 11
    BY_TIME_ELEMENT = 12
    FROM_INITIAL_CONDITIONS = 13
    SETPOINT = 14


class ElementInitializationInputOption(enum.Enum):
    SETPOINT = 21
    INITIAL = 22
    CURRENT_VALUES = 23


class InputOption(enum.Enum):
    CURRENT = 20
    INITIAL = 21
    SETPOINT = 22
    # TODO: PREVIOUS = 23


class TimeResolutionOption(enum.Enum):
    COLLOCATION_POINTS = 31
    FINITE_ELEMENTS = 32
    SAMPLE_POINTS = 33
    INITIAL_POINT = 34


class ControlPenaltyType(enum.Enum):
    ERROR = 41
    ACTION = 42
    NONE = 43


class VariableCategory(enum.Enum):
    DIFFERENTIAL = 1
    ALGEBRAIC = 2
    DERIVATIVE = 3
    INPUT = 4
    FIXED = 5
    SCALAR = 6
    UNUSED = 7
    DISTURBANCE = 8
    MEASUREMENT = 9


class ConstraintCategory(enum.Enum):
    DIFFERENTIAL = 1
    ALGEBRAIC = 2
    DISCRETIZATION = 3
    INPUT = 4
    INITIAL = 5
    TERMINAL = 6
    SENSITIVITY = 7
    SCALAR = 8
    UNUSED = 9
    INEQUALITY = 10


class NoiseBoundOption(enum.Enum):
    FAIL = 60
    DISCARD = 61
    PUSH = 62


class PlantHorizonType(enum.Enum):
    FULL = 71
    ROLLING = 72


# This function is used as the domain for the user-provided
# list of inputs at time.first().
def validate_list_of_vardata(varlist):
    if not isinstance(varlist, list):
        raise TypeError("Not a list of VarData")
    for var in varlist:
        if not isinstance(var, VarData):
            raise TypeError("Not a list of VarData")
    return varlist


def validate_list_of_vardata_value_tuples(varvaluelist):
    if not isinstance(varvaluelist, list):
        raise TypeError("Not a list")
    for item in varvaluelist:
        if not isinstance(item, tuple):
            raise TypeError("Item in list is not a tuple")
        if not len(item) == 2:
            raise ValueError("Tuple in list does not have correct length")
        if not isinstance(item[0], VarData):
            raise TypeError("First entry is not a VarData")
        item = (item[0], float(item[1]))
    return varvaluelist


def validate_solver(solver):
    if type(solver) is str:
        solver = SolverFactory(solver)
    if not hasattr(solver, "solve"):
        raise TypeError("Solver does not implement solve method")
    return solver


def get_ncp(continuous_set):
    scheme = continuous_set.get_discretization_info()["scheme"]
    if scheme == "LAGRANGE-RADAU":
        return continuous_set.get_discretization_info()["ncp"]
    elif scheme == "BACKWARD Difference":
        return 1
    else:
        raise NotImplementedError("%s discretization scheme is not supported" % scheme)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This module contains miscellaneous utility functions for use in IDAES models.
"""

from io import StringIO
import pytest
import re

from pyomo.environ import ConcreteModel, Constraint, Expression, Set, Block, Var, units
from pyomo.common.config import ConfigBlock
from pyomo.core.base.units_container import UnitsError

from idaes.core.util.misc import (
    add_object_reference,
    set_param_from_config,
    compact_expression_to_string,
    print_compact_form,
)
import idaes.logger as idaeslog


# Author: Andrew Lee
@pytest.mark.unit
def test_add_object_reference():
    m = ConcreteModel()

    m.s = Set(initialize=[1, 2, 3])

    add_object_reference(m, "test_ref", m.s)

    assert hasattr(m, "test_ref")
    assert m.test_ref == m.s


# Author: Andrew Lee
@pytest.mark.unit
def test_add_object_reference_fail():
    m = ConcreteModel()

    with pytest.raises(AttributeError):
        add_object_reference(m, "test_ref", m.s)


@pytest.mark.unit
class TestSetParamFromConfig:
    def test_default_config(self, caplog):
        caplog.set_level(idaeslog.DEBUG, logger=("idaes.core.util.misc"))

        m = ConcreteModel()
        m.b = Block()
        m.b.config = ConfigBlock(implicit=True)
        m.b.config.parameter_data = {"test_param": 42}

        m.b.test_param = Var(initialize=1)

        set_param_from_config(m.b, "test_param")

        assert m.b.test_param.value == 42

        assert (
            "b no units provided for parameter test_param - assuming "
            "default units" in caplog.text
        )

    def test_specified_config(self):
        m = ConcreteModel()
        m.b = Block()
        m.b.config2 = ConfigBlock(implicit=True)
        m.b.config2.parameter_data = {"test_param": 42}

        m.b.test_param = Var(initialize=1)

        set_param_from_config(m.b, "test_param", config=m.b.config2)

        assert m.b.test_param.value == 42

    def test_no_config(self):
        m = ConcreteModel()
        m.b = Block()

        m.b.test_param = Var(initialize=1)

        with pytest.raises(
            AttributeError,
            match="b - set_param_from_config method was "
            "not provided with a config argument, but no "
            "default Config block exists. Please specify the "
            "Config block to use via the config argument.",
        ):
            set_param_from_config(m.b, "test_param")

    def test_invalid_config(self):
        m = ConcreteModel()
        m.b = Block()
        m.b.config = "foo"

        m.b.test_param = Var(initialize=1)

        with pytest.raises(
            TypeError,
            match="b - set_param_from_config - config argument "
            "provided is not an instance of a Config Block.",
        ):
            set_param_from_config(m.b, "test_param")

    def test_no_param(self):
        m = ConcreteModel()
        m.b = Block()
        m.b.config = ConfigBlock(implicit=True)
        m.b.config.parameter_data = {"test_param": 42}

        with pytest.raises(
            AttributeError,
            match="b - set_param_from_config method was "
            "provided with param argument test_param, but no "
            "attribute of that name exists.",
        ):
            set_param_from_config(m.b, "test_param")

    def test_no_parameter_data(self):
        m = ConcreteModel()
        m.b = Block()
        m.b.config = ConfigBlock(implicit=True)
        m.b.config.parameter_data = {}

        m.b.test_param = Var(initialize=1)

        with pytest.raises(
            KeyError,
            match="b - set_param_from_config method was "
            "provided with param argument test_param, but the "
            "config block does not contain a value for this "
            "parameter.",
        ):
            set_param_from_config(m.b, "test_param")

    def test_indexed(self):
        m = ConcreteModel()
        m.b = Block()
        m.b.config = ConfigBlock(implicit=True)
        m.b.config.parameter_data = {"test_param": {"1": 42}}

        m.b.test_param_1 = Var(initialize=1)

        set_param_from_config(m.b, "test_param", index="1")

        assert m.b.test_param_1.value == 42

    def test_no_param_indexed(self):
        m = ConcreteModel()
        m.b = Block()
        m.b.config = ConfigBlock(implicit=True)
        m.b.config.parameter_data = {"test_param": {"1": 42}}

        m.b.test_param = Var(initialize=1)

        with pytest.raises(
            AttributeError,
            match=re.escape(
                "b - set_param_from_config method was "
                "provided with param and index arguments "
                "test_param 1, but no attribute with that "
                "combination (test_param_1) exists."
            ),
        ):
            set_param_from_config(m.b, "test_param", index="1")

    def test_no_parameter_data_indexed(self):
        m = ConcreteModel()
        m.b = Block()
        m.b.config = ConfigBlock(implicit=True)
        m.b.config.parameter_data = {"test_param": {"2": 42}}

        m.b.test_param_1 = Var(initialize=1)

        with pytest.raises(
            KeyError,
            match="b - set_param_from_config method was "
            "provided with param and index arguments "
            "test_param 1, but the config block does not "
            "contain a value for this parameter and index.",
        ):
            set_param_from_config(m.b, "test_param", index="1")

    def test_dimensionless_default(self, caplog):
        caplog.set_level(idaeslog.DEBUG, logger=("idaes.core.util.misc"))

        m = ConcreteModel()
        m.b = Block()
        m.b.config = ConfigBlock(implicit=True)
        m.b.config.parameter_data = {"test_param": 42}

        m.b.test_param = Var(initialize=1, units=units.dimensionless)

        set_param_from_config(m.b, "test_param")

        assert m.b.test_param.value == 42

        assert (
            "b no units provided for parameter test_param - assuming "
            "default units" in caplog.text
        )

    def test_dimensionless_defined(self, caplog):
        caplog.set_level(idaeslog.DEBUG, logger=("idaes.core.util.misc"))

        m = ConcreteModel()
        m.b = Block()
        m.b.config = ConfigBlock(implicit=True)
        m.b.config.parameter_data = {"test_param": (42, units.dimensionless)}

        m.b.test_param = Var(initialize=1, units=units.dimensionless)

        set_param_from_config(m.b, "test_param")

        assert m.b.test_param.value == 42

    def test_dimensionless_defined_none(self, caplog):
        caplog.set_level(idaeslog.DEBUG, logger=("idaes.core.util.misc"))

        m = ConcreteModel()
        m.b = Block()
        m.b.config = ConfigBlock(implicit=True)
        m.b.config.parameter_data = {"test_param": (42, None)}

        m.b.test_param = Var(initialize=1, units=units.dimensionless)

        set_param_from_config(m.b, "test_param")

        assert m.b.test_param.value == 42

    def test_none_defined(self, caplog):
        caplog.set_level(idaeslog.DEBUG, logger=("idaes.core.util.misc"))

        m = ConcreteModel()
        m.b = Block()
        m.b.config = ConfigBlock(implicit=True)
        m.b.config.parameter_data = {"test_param": (42, None)}

        m.b.test_param = Var(initialize=1, units=None)

        set_param_from_config(m.b, "test_param")

        assert m.b.test_param.value == 42

    def test_none_defined_dimensionless(self, caplog):
        caplog.set_level(idaeslog.DEBUG, logger=("idaes.core.util.misc"))

        m = ConcreteModel()
        m.b = Block()
        m.b.config = ConfigBlock(implicit=True)
        m.b.config.parameter_data = {"test_param": (42, units.dimensionless)}

        m.b.test_param = Var(initialize=1, units=None)

        set_param_from_config(m.b, "test_param")

        assert m.b.test_param.value == 42

    def test_consistent_units(self, caplog):
        caplog.set_level(idaeslog.DEBUG, logger=("idaes.core.util.misc"))

        m = ConcreteModel()
        m.b = Block()
        m.b.config = ConfigBlock(implicit=True)
        m.b.config.parameter_data = {"test_param": (42, units.m)}

        m.b.test_param = Var(initialize=1, units=units.m)

        set_param_from_config(m.b, "test_param")

        assert m.b.test_param.value == 42

    def test_inconsistent_units(self, caplog):
        caplog.set_level(idaeslog.DEBUG, logger=("idaes.core.util.misc"))

        m = ConcreteModel()
        m.b = Block()
        m.b.config = ConfigBlock(implicit=True)
        m.b.config.parameter_data = {"test_param": (42, units.m)}

        m.b.test_param = Var(initialize=1, units=units.s)

        with pytest.raises(
            UnitsError, match="Cannot convert m to s. Units are not compatible."
        ):
            set_param_from_config(m.b, "test_param")

    def test_inconsistent_units_dimensionless(self, caplog):
        caplog.set_level(idaeslog.DEBUG, logger=("idaes.core.util.misc"))

        m = ConcreteModel()
        m.b = Block()
        m.b.config = ConfigBlock(implicit=True)
        m.b.config.parameter_data = {"test_param": (42, units.dimensionless)}

        m.b.test_param = Var(initialize=1, units=units.s)

        with pytest.raises(
            UnitsError,
            match="Cannot convert dimensionless to s. Units are not compatible.",
        ):
            set_param_from_config(m.b, "test_param")

    def test_inconsistent_units_none(self, caplog):
        caplog.set_level(idaeslog.DEBUG, logger=("idaes.core.util.misc"))

        m = ConcreteModel()
        m.b = Block()
        m.b.config = ConfigBlock(implicit=True)
        m.b.config.parameter_data = {"test_param": (42, None)}

        m.b.test_param = Var(initialize=1, units=units.s)

        with pytest.raises(
            UnitsError,
            match="Cannot convert dimensionless to s. Units are not compatible.",
        ):
            set_param_from_config(m.b, "test_param")

    def test_inconsistent_units_dimensionless_2(self, caplog):
        caplog.set_level(idaeslog.DEBUG, logger=("idaes.core.util.misc"))

        m = ConcreteModel()
        m.b = Block()
        m.b.config = ConfigBlock(implicit=True)
        m.b.config.parameter_data = {"test_param": (42, units.s)}

        m.b.test_param = Var(initialize=1, units=units.dimensionless)

        with pytest.raises(
            UnitsError,
            match="Cannot convert s to dimensionless. Units are not compatible.",
        ):
            set_param_from_config(m.b, "test_param")

    def test_unitted_default(self, caplog):
        caplog.set_level(idaeslog.DEBUG, logger=("idaes.core.util.misc"))

        m = ConcreteModel()
        m.b = Block()
        m.b.config = ConfigBlock(implicit=True)
        m.b.config.parameter_data = {"test_param": 42}

        m.b.test_param = Var(initialize=1, units=units.m)

        set_param_from_config(m.b, "test_param")

        assert m.b.test_param.value == 42

        assert (
            "b no units provided for parameter test_param - assuming "
            "default units" in caplog.text
        )


class TestToExprStringVisitor:
    @pytest.mark.unit
    def test_compact_expression_to_string(self):
        m = ConcreteModel()
        m.v1 = Var()
        m.v2 = Var()
        m.v3 = Var()

        m.e1 = Expression(expr=m.v1 + m.v2)
        m.c1 = Constraint(expr=0 == m.v3 * m.e1)

        str = compact_expression_to_string(m.c1.expr)
        expected = "v3*e1  ==  0"

        assert str == expected

    @pytest.mark.unit
    def test_print_compact_form_expr(self):
        m = ConcreteModel()
        m.v1 = Var()
        m.v2 = Var()
        m.v3 = Var()

        m.e1 = Expression(expr=m.v1 + m.v2)
        m.c1 = Constraint(expr=0 == m.v3 * m.e1)

        expected = "v3*e1  ==  0"

        stream = StringIO()
        print_compact_form(m.c1.expr, stream=stream)

        assert stream.getvalue() == expected

    @pytest.mark.unit
    def test_print_compact_form_component(self):
        m = ConcreteModel()
        m.v1 = Var()
        m.v2 = Var()
        m.v3 = Var()

        m.e1 = Expression(expr=m.v1 + m.v2)
        m.c1 = Constraint(expr=0 == m.v3 * m.e1)

        expected = "v3*e1  ==  0"

        stream = StringIO()
        print_compact_form(m.c1, stream=stream)

        assert stream.getvalue() == expected



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Example for Caprese's module for NMPC.
"""
import random
from idaes.apps.caprese.nmpc import NMPCSim
from idaes.apps.caprese.dynamic_block import DynamicBlock
from idaes.apps.caprese.controller import ControllerBlock
from idaes.apps.caprese.util import apply_noise_with_bounds
from idaes.apps.caprese.categorize import (
    categorize_dae_variables_and_constraints,
    VariableCategory,
    ConstraintCategory,
)
from idaes.core.solvers import get_solver

VC = VariableCategory
CC = ConstraintCategory

import pyomo.environ as pyo
from pyomo.dae.flatten import flatten_dae_components
from pyomo.dae.initialization import solve_consistent_initial_conditions
from pyomo.contrib.pynumero.interfaces.pyomo_nlp import PyomoNLP
from pyomo.contrib.incidence_analysis.interface import IncidenceGraphInterface
from pyomo.core.expr.calculus.derivatives import reverse_ad

import idaes.logger as idaeslog
from idaes.apps.caprese.examples.cstr_model import make_model

import numpy as np
import scipy.sparse as sps
import pandas as pd
import matplotlib.pyplot as plt

__author__ = "Robert Parker"


# See if ipopt is available and set up solver
if pyo.SolverFactory("ipopt").available():
    solver = get_solver(
        solver="ipopt",
        options={
            "tol": 1e-6,
            "bound_push": 1e-8,
            "halt_on_ampl_error": "yes",
            "linear_solver": "ma57",
        },
    )
else:
    solver = None


class PlotData(object):
    def __init__(self, group, location, name=None, t_switch=None):
        # Would really like a PlotData class that is constructed based on an
        # NMPCVar object that contains necessary setpoint/reference
        # information, instead of having to access that in the NMPCVarGroup
        time = group.index_set
        if t_switch == None:
            t_switch = group.t0

        self.name = name

        var = group.varlist[location]
        initial = group.reference[location]
        setpoint = group.setpoint[location]
        self.data_series = pd.Series(
            [var[t].value for t in time], index=[t for t in time]
        )
        self.setpoint_series = pd.Series(
            [initial if t < t_switch else setpoint for t in time]
        )

    def plot(self):
        # fig, ax can be formatted to the user's liking
        fig, ax = plt.subplots()
        if self.name is not None:
            self.data_series.plot(label=self.name)
        else:
            self.data_series.plot()
        return fig, ax


def main(plot_switch=False):

    # This tests the same model constructed in the test_nmpc_constructor_1 file
    m_controller = make_model(horizon=3, ntfe=30, ntcp=2, bounds=True)
    sample_time = 0.5
    m_plant = make_model(horizon=sample_time, ntfe=5, ntcp=2)
    time_plant = m_plant.fs.time

    solve_consistent_initial_conditions(m_plant, time_plant, solver)

    #####
    # Flatten and categorize controller model
    #####
    model = m_controller
    time = model.fs.time
    t0 = time.first()
    t1 = time[2]
    scalar_vars, dae_vars = flatten_dae_components(
        model,
        time,
        pyo.Var,
    )
    scalar_cons, dae_cons = flatten_dae_components(
        model,
        time,
        pyo.Constraint,
    )
    inputs = [
        model.fs.mixer.S_inlet.flow_vol,
        model.fs.mixer.E_inlet.flow_vol,
    ]
    measurements = [
        pyo.Reference(model.fs.cstr.outlet.conc_mol[:, "C"]),
        pyo.Reference(model.fs.cstr.outlet.conc_mol[:, "E"]),
        pyo.Reference(model.fs.cstr.outlet.conc_mol[:, "S"]),
        pyo.Reference(model.fs.cstr.outlet.conc_mol[:, "P"]),
        model.fs.cstr.outlet.temperature,
    ]
    model.fs.cstr.control_volume.material_holdup[:, "aq", "Solvent"].fix()
    model.fs.cstr.total_flow_balance.deactivate()
    var_partition, con_partition = categorize_dae_variables_and_constraints(
        model,
        dae_vars,
        dae_cons,
        time,
        input_vars=inputs,
    )
    controller = ControllerBlock(
        model=model,
        time=time,
        measurements=measurements,
        category_dict={None: var_partition},
    )
    controller.construct()

    solve_consistent_initial_conditions(m_controller, time, solver)
    controller.initialize_to_initial_conditions()

    m_controller._dummy_obj = pyo.Objective(expr=0)
    nlp = PyomoNLP(m_controller)
    igraph = IncidenceGraphInterface(nlp)
    m_controller.del_component(m_controller._dummy_obj)
    diff_vars = [var[t1] for var in var_partition[VC.DIFFERENTIAL]]
    alg_vars = [var[t1] for var in var_partition[VC.ALGEBRAIC]]
    deriv_vars = [var[t1] for var in var_partition[VC.DERIVATIVE]]
    diff_eqns = [con[t1] for con in con_partition[CC.DIFFERENTIAL]]
    alg_eqns = [con[t1] for con in con_partition[CC.ALGEBRAIC]]

    # Assemble and factorize "derivative Jacobian"
    dfdz = nlp.extract_submatrix_jacobian(diff_vars, diff_eqns)
    dfdy = nlp.extract_submatrix_jacobian(alg_vars, diff_eqns)
    dgdz = nlp.extract_submatrix_jacobian(diff_vars, alg_eqns)
    dgdy = nlp.extract_submatrix_jacobian(alg_vars, alg_eqns)
    dfdzdot = nlp.extract_submatrix_jacobian(deriv_vars, diff_eqns)
    fact = sps.linalg.splu(dgdy.tocsc())
    dydz = fact.solve(dgdz.toarray())
    deriv_jac = dfdz - dfdy.dot(dydz)
    fact = sps.linalg.splu(dfdzdot.tocsc())
    dzdotdz = -fact.solve(deriv_jac)

    # Use some heuristic on the eigenvalues of the derivative Jacobian
    # to identify fast states.
    w, V = np.linalg.eig(dzdotdz)
    w_max = np.max(np.abs(w))
    (fast_modes,) = np.where(np.abs(w) > w_max / 2)
    fast_states = []
    for idx in fast_modes:
        evec = V[:, idx]
        _fast_states, _ = np.where(np.abs(evec) > 0.5)
        fast_states.extend(_fast_states)
    fast_states = set(fast_states)

    # Store components necessary for model reduction in a model-
    # independent form.
    fast_state_derivs = [
        pyo.ComponentUID(var_partition[VC.DERIVATIVE][idx].referent, context=model)
        for idx in fast_states
    ]
    fast_state_diffs = [
        pyo.ComponentUID(var_partition[VC.DIFFERENTIAL][idx].referent, context=model)
        for idx in fast_states
    ]
    fast_state_discs = [
        pyo.ComponentUID(con_partition[CC.DISCRETIZATION][idx].referent, context=model)
        for idx in fast_states
    ]

    # Perform pseudo-steady state model reduction on the fast states
    # and re-categorize
    for cuid in fast_state_derivs:
        var = cuid.find_component_on(m_controller)
        var.fix(0.0)
    for cuid in fast_state_diffs:
        var = cuid.find_component_on(m_controller)
        var[t0].unfix()
    for cuid in fast_state_discs:
        con = cuid.find_component_on(m_controller)
        con.deactivate()

    var_partition, con_partition = categorize_dae_variables_and_constraints(
        model,
        dae_vars,
        dae_cons,
        time,
        input_vars=inputs,
    )
    controller.del_component(model)

    # Re-construct controller block with new categorization
    measurements = [
        pyo.Reference(model.fs.cstr.outlet.conc_mol[:, "C"]),
        pyo.Reference(model.fs.cstr.outlet.conc_mol[:, "E"]),
        pyo.Reference(model.fs.cstr.outlet.conc_mol[:, "S"]),
        pyo.Reference(model.fs.cstr.outlet.conc_mol[:, "P"]),
    ]
    controller = ControllerBlock(
        model=model,
        time=time,
        measurements=measurements,
        category_dict={None: var_partition},
    )
    controller.construct()

    #####
    # Construct dynamic block for plant
    #####
    model = m_plant
    time = model.fs.time
    t0 = time.first()
    t1 = time[2]
    scalar_vars, dae_vars = flatten_dae_components(
        model,
        time,
        pyo.Var,
    )
    scalar_cons, dae_cons = flatten_dae_components(
        model,
        time,
        pyo.Constraint,
    )
    inputs = [
        model.fs.mixer.S_inlet.flow_vol,
        model.fs.mixer.E_inlet.flow_vol,
    ]
    measurements = [
        pyo.Reference(model.fs.cstr.outlet.conc_mol[:, "C"]),
        pyo.Reference(model.fs.cstr.outlet.conc_mol[:, "E"]),
        pyo.Reference(model.fs.cstr.outlet.conc_mol[:, "S"]),
        pyo.Reference(model.fs.cstr.outlet.conc_mol[:, "P"]),
    ]
    model.fs.cstr.control_volume.material_holdup[:, "aq", "Solvent"].fix()
    model.fs.cstr.total_flow_balance.deactivate()

    var_partition, con_partition = categorize_dae_variables_and_constraints(
        model,
        dae_vars,
        dae_cons,
        time,
        input_vars=inputs,
    )
    plant = DynamicBlock(
        model=model,
        time=time,
        measurements=measurements,
        category_dict={None: var_partition},
    )
    plant.construct()

    p_t0 = plant.time.first()
    c_t0 = controller.time.first()
    p_ts = plant.sample_points[1]
    c_ts = controller.sample_points[1]

    controller.set_sample_time(sample_time)
    plant.set_sample_time(sample_time)

    # We now perform the "RTO" calculation: Find the optimal steady state
    # to achieve the following setpoint
    setpoint = [
        (controller.mod.fs.cstr.outlet.conc_mol[0, "P"], 0.4),
        # (controller.mod.fs.cstr.outlet.conc_mol[0, 'S'], 0.01),
        (controller.mod.fs.cstr.outlet.conc_mol[0, "S"], 0.1),
        (controller.mod.fs.cstr.control_volume.energy_holdup[0, "aq"], 300),
        (controller.mod.fs.mixer.E_inlet.flow_vol[0], 0.1),
        (controller.mod.fs.mixer.S_inlet.flow_vol[0], 2.0),
        (controller.mod.fs.cstr.volume[0], 1.0),
    ]
    setpoint_weights = [
        (controller.mod.fs.cstr.outlet.conc_mol[0, "P"], 1.0),
        (controller.mod.fs.cstr.outlet.conc_mol[0, "S"], 1.0),
        (controller.mod.fs.cstr.control_volume.energy_holdup[0, "aq"], 1.0),
        (controller.mod.fs.mixer.E_inlet.flow_vol[0], 1.0),
        (controller.mod.fs.mixer.S_inlet.flow_vol[0], 1.0),
        (controller.mod.fs.cstr.volume[0], 1.0),
    ]

    # Some of the "differential variables" that have been fixed in the
    # model file are different from the measurements listed above. We
    # unfix them here so the RTO solve is not overconstrained.
    # (The RTO solve will only automatically unfix inputs and measurements.)
    controller.mod.fs.cstr.control_volume.material_holdup[0, ...].unfix()
    controller.mod.fs.cstr.control_volume.energy_holdup[0, ...].unfix()
    # controller.mod.fs.cstr.volume[0].unfix()
    controller.mod.fs.cstr.control_volume.material_holdup[0, "aq", "Solvent"].fix()

    controller.add_setpoint_objective(setpoint, setpoint_weights)
    controller.solve_setpoint(solver)

    # Now we are ready to construct the tracking NMPC problem
    tracking_weights = [
        *((v, 1.0) for v in controller.vectors.differential[:, 0]),
        *((v, 1.0) for v in controller.vectors.input[:, 0]),
    ]

    controller.add_tracking_objective(tracking_weights)

    controller.constrain_control_inputs_piecewise_constant()

    controller.initialize_to_initial_conditions()

    # Solve the first control problem
    controller.vectors.input[...].unfix()
    controller.vectors.input[:, 0].fix()
    solver.solve(controller, tee=True)

    # For a proper NMPC simulation, we must have noise.
    # We do this by treating inputs and measurements as Gaussian random
    # variables with the following variances (and bounds).
    cstr = controller.mod.fs.cstr
    variance = [
        (cstr.outlet.conc_mol[0.0, "S"], 0.01),
        (cstr.outlet.conc_mol[0.0, "E"], 0.005),
        (cstr.outlet.conc_mol[0.0, "C"], 0.01),
        (cstr.outlet.conc_mol[0.0, "P"], 0.005),
        (cstr.outlet.temperature[0.0], 1.0),
        (cstr.volume[0.0], 0.05),
    ]
    controller.set_variance(variance)
    measurement_variance = [v.variance for v in controller.MEASUREMENT_BLOCK[:].var]
    measurement_noise_bounds = [
        (0.0, var[c_t0].ub) for var in controller.MEASUREMENT_BLOCK[:].var
    ]

    mx = plant.mod.fs.mixer
    variance = [
        (mx.S_inlet_state[0.0].flow_vol, 0.02),
        (mx.E_inlet_state[0.0].flow_vol, 0.001),
    ]
    plant.set_variance(variance)
    input_variance = [v.variance for v in plant.INPUT_BLOCK[:].var]
    input_noise_bounds = [(0.0, var[p_t0].ub) for var in plant.INPUT_BLOCK[:].var]

    random.seed(100)

    # Extract inputs from controller and inject them into plant
    inputs = controller.generate_inputs_at_time(c_ts)
    plant.inject_inputs(inputs)

    # This "initialization" really simulates the plant with the new inputs.
    plant.vectors.input[:, :].fix()
    plant.initialize_by_solving_elements(solver)
    plant.vectors.input[:, :].fix()
    solver.solve(plant, tee=True)

    for i in range(1, 11):
        print("\nENTERING NMPC LOOP ITERATION %s\n" % i)
        measured = plant.generate_measurements_at_time(p_ts)
        plant.advance_one_sample()
        plant.initialize_to_initial_conditions()
        measured = apply_noise_with_bounds(
            measured,
            measurement_variance,
            random.gauss,
            measurement_noise_bounds,
        )

        controller.advance_one_sample()
        controller.load_measurements(measured)

        solver.solve(controller, tee=True)

        inputs = controller.generate_inputs_at_time(c_ts)
        inputs = apply_noise_with_bounds(
            inputs,
            input_variance,
            random.gauss,
            input_noise_bounds,
        )
        plant.inject_inputs(inputs)

        plant.initialize_by_solving_elements(solver)
        solver.solve(plant)

    import pdb

    pdb.set_trace()


if __name__ == "__main__":
    main()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
1D Single pass shell and tube HX model with 0D wall conduction model.

This model derives from the HeatExchanger1D unit model.
"""
# Import Pyomo libraries
from pyomo.environ import (
    Block,
    Var,
    check_optimal_termination,
    Constraint,
    value,
    units as pyunits,
)
from pyomo.common.config import ConfigValue, Bool

# Import IDAES cores
from idaes.core import declare_process_block_class, UnitModelBlockData
from idaes.models.unit_models.heat_exchanger_1D import (
    HeatExchanger1DData,
)
from idaes.core.util.misc import add_object_reference
from idaes.core.util.exceptions import InitializationError
from idaes.core.util.constants import Constants as c
from idaes.core.util import scaling as iscale
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.solvers import get_solver
from idaes.core.initialization import SingleControlVolumeUnitInitializer
import idaes.logger as idaeslog


__author__ = "Jaffer Ghouse"

# Set up logger
_log = idaeslog.getLogger(__name__)


class ShellAndTubeInitializer(SingleControlVolumeUnitInitializer):
    """
    Initializer for 1D Shell and Tube Heat Exchanger units.

    """

    def initialization_routine(
        self,
        model: Block,
        plugin_initializer_args: dict = None,
    ):
        """
        Common initialization routine for 1D Shell and Tube Heat Exchangers.

        This routine starts by initializing the hot and cold side properties. Next, the hot side is solved with
        the wall temperature fixed to the average of the hot and cold side temperatures and the heat
        transfer constraints deactivated. Finally, full model is solved with the wall temperature unfixed.

        Args:
            model: Pyomo Block to be initialized
            plugin_initializer_args: dict-of-dicts containing arguments to be passed to plug-in Initializers.
                Keys should be submodel components.

        Returns:
            Pyomo solver results object
        """
        return super(SingleControlVolumeUnitInitializer, self).initialization_routine(
            model=model,
            plugin_initializer_args=plugin_initializer_args,
        )

    def initialize_main_model(
        self,
        model: Block,
    ):
        """
        Initialization routine for main 1D Shell and Tube HX models.

        Args:
            model: Pyomo Block to be initialized.

        Returns:
            Pyomo solver results object.

        """
        init_log = idaeslog.getInitLogger(
            model.name, self.get_output_level(), tag="unit"
        )
        solve_log = idaeslog.getSolveLogger(
            model.name, self.get_output_level(), tag="unit"
        )

        # Create solver
        solver = self._get_solver()

        # ---------------------------------------------------------------------
        # Initialize control volumes
        self.initialize_control_volume(model.hot_side)
        self.initialize_control_volume(model.cold_side)

        init_log.info_high("Initialization Step 1 Complete.")

        # ---------------------------------------------------------------------
        # Solve hot side
        hot_side_units = (
            model.hot_side.config.property_package.get_metadata().get_derived_units
        )
        for t in model.flowsheet().time:
            for z in model.hot_side.length_domain:
                model.temperature_wall[t, z].fix(
                    0.5
                    * (
                        model.hot_side.properties[t, 0].temperature
                        + pyunits.convert(
                            model.cold_side.properties[t, 0].temperature,
                            to_units=hot_side_units("temperature"),
                        )
                    )
                )

        model.cold_side.deactivate()
        model.cold_side_heat_transfer_eq.deactivate()
        model.heat_conservation.deactivate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solver.solve(model, tee=slc.tee)
        init_log.info_high(f"Initialization Step 2 {idaeslog.condition(res)}.")

        # ---------------------------------------------------------------------
        # Solve full unit
        model.cold_side.activate()
        model.cold_side_heat_transfer_eq.activate()
        model.heat_conservation.activate()
        model.temperature_wall.unfix()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solver.solve(model, tee=slc.tee)
        init_log.info_high(f"Initialization Step 3 {idaeslog.condition(res)}.")

        return res


@declare_process_block_class("ShellAndTube1D")
class ShellAndTube1DData(HeatExchanger1DData):
    """1D Shell and Tube HX Unit Model Class."""

    default_initializer = ShellAndTubeInitializer

    CONFIG = HeatExchanger1DData.CONFIG(implicit=True)
    CONFIG.declare(
        "shell_is_hot",
        ConfigValue(
            default=True,
            domain=Bool,
            description="Shell side contains hot fluid",
            doc="""Boolean flag indicating whether shell side contains hot fluid (default=True).
    If True, shell side will be the hot_side, if False shell side will be cold_side.""",
        ),
    )

    def _process_config(self):
        super()._process_config()

        # Check for custom names, and if not present assign defaults
        if self.config.hot_side_name is None:
            if self.config.shell_is_hot:
                self.config.hot_side_name = "Shell"
            else:
                self.config.hot_side_name = "Tube"

        if self.config.cold_side_name is None:
            if self.config.shell_is_hot:
                self.config.cold_side_name = "Tube"
            else:
                self.config.cold_side_name = "Shell"

    def _make_geometry(self):
        # Add reference to control volume geometry
        add_object_reference(self, "hot_side_area", self.hot_side.area)
        add_object_reference(self, "length", self.hot_side.length)

        # Equate hot and cold side geometries
        hot_side_units = (
            self.hot_side.config.property_package.get_metadata().get_derived_units
        )
        cold_side_units = (
            self.cold_side.config.property_package.get_metadata().get_derived_units
        )

        @self.Constraint(
            self.flowsheet().time, doc="Equating hot and cold side lengths"
        )
        def length_equality(self, t):
            return (
                pyunits.convert(
                    self.cold_side.length, to_units=hot_side_units("length")
                )
                == self.hot_side.length
            )

        # Get hot and cold sides
        if self.config.shell_is_hot:
            shell = self.hot_side
            tube = self.cold_side
        else:
            shell = self.cold_side
            tube = self.hot_side

        # New Unit model variables and constraints
        self.shell_diameter = Var(
            initialize=0.011,
            doc="Diameter of shell",
            units=hot_side_units("length"),
        )
        self.tube_outer_diameter = Var(
            initialize=1,
            doc="Outer diameter of tubes",
            units=hot_side_units("length"),
        )
        self.tube_inner_diameter = Var(
            initialize=0.010,
            doc="Inner diameter of tubes",
            units=hot_side_units("length"),
        )
        self.number_of_tubes = Var(
            initialize=1, doc="Number of tubes", units=pyunits.dimensionless
        )

        # Calculate cross-sectional area of control volumes
        self.tube_side_xsec_area_calc = Constraint(
            expr=4 * tube.area
            == self.number_of_tubes
            * c.pi
            * pyunits.convert(
                self.tube_inner_diameter, to_units=cold_side_units("length")
            )
            ** 2,
            doc="Tube side cross-sectional area",
        )
        # Need to account for area occupied by tubes
        self.shell_side_xsec_area_calc = Constraint(
            expr=4 * shell.area
            == c.pi
            * (
                self.shell_diameter**2
                - self.number_of_tubes * self.tube_outer_diameter**2
            ),
            doc="Shell side cross-sectional area",
        )

    def _make_performance(self):
        hot_side_units = (
            self.hot_side.config.property_package.get_metadata().get_derived_units
        )

        # Performance variables
        self.hot_side_heat_transfer_coefficient = Var(
            self.flowsheet().time,
            self.hot_side.length_domain,
            initialize=50,
            doc="Hot side heat transfer coefficient",
            units=hot_side_units("heat_transfer_coefficient"),
        )
        self.cold_side_heat_transfer_coefficient = Var(
            self.flowsheet().time,
            self.cold_side.length_domain,
            initialize=50,
            doc="Cold side heat transfer coefficient",
            units=hot_side_units("heat_transfer_coefficient"),
        )
        self.temperature_wall = Var(
            self.flowsheet().time,
            self.hot_side.length_domain,
            initialize=298.15,
            units=hot_side_units("temperature"),
        )

        @self.Constraint(
            self.flowsheet().time,
            self.hot_side.length_domain,
            doc="Heat transfer between hot_side and wall",
        )
        def hot_side_heat_transfer_eq(self, t, x):
            return self.hot_side.heat[t, x] == -(
                self.hot_side_heat_transfer_coefficient[t, x]
                * self.number_of_tubes
                * c.pi
                * self.tube_outer_diameter
                * (
                    self.hot_side.properties[t, x].temperature
                    - self.temperature_wall[t, x]
                )
            )

        @self.Constraint(
            self.flowsheet().time,
            self.cold_side.length_domain,
            doc="Heat transfer between cold_side and wall",
        )
        def cold_side_heat_transfer_eq(self, t, x):
            return pyunits.convert(
                self.cold_side.heat[t, x],
                to_units=hot_side_units("power") / hot_side_units("length"),
            ) == (
                self.cold_side_heat_transfer_coefficient[t, x]
                * self.number_of_tubes
                * c.pi
                * self.tube_inner_diameter
                * (
                    self.temperature_wall[t, x]
                    - pyunits.convert(
                        self.cold_side.properties[t, x].temperature,
                        to_units=hot_side_units("temperature"),
                    )
                )
            )

    def initialize_build(
        self,
        hot_side_state_args=None,
        cold_side_state_args=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        Initialization routine for the unit.

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialize hot_side block
        flags_hot_side = self.hot_side.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=hot_side_state_args,
        )

        flags_cold_side = self.cold_side.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=cold_side_state_args,
        )

        init_log.info_high("Initialization Step 1 Complete.")

        # ---------------------------------------------------------------------
        # Solve unit
        hot_side_units = (
            self.hot_side.config.property_package.get_metadata().get_derived_units
        )
        for t in self.flowsheet().time:
            for z in self.hot_side.length_domain:
                self.temperature_wall[t, z].fix(
                    value(
                        0.5
                        * (
                            self.hot_side.properties[t, 0].temperature
                            + pyunits.convert(
                                self.cold_side.properties[t, 0].temperature,
                                to_units=hot_side_units("temperature"),
                            )
                        )
                    )
                )

        self.cold_side.deactivate()
        self.cold_side_heat_transfer_eq.deactivate()
        self.heat_conservation.deactivate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(res)))

        self.cold_side.activate()
        self.cold_side_heat_transfer_eq.activate()
        self.heat_conservation.activate()
        self.temperature_wall.unfix()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 3 {}.".format(idaeslog.condition(res)))

        self.hot_side.release_state(flags_hot_side)
        self.cold_side.release_state(flags_cold_side)

        if res is not None and not check_optimal_termination(res):
            raise InitializationError(
                f"{self.name} failed to initialize successfully. Please check "
                f"the output logs for more information."
            )

        init_log.info("Initialization Complete.")

    def _get_performance_contents(self, time_point=0):
        var_dict = {}
        var_dict["Shell Diameter"] = self.shell_diameter
        var_dict["Length"] = self.length
        var_dict["Tube Outer Diameter"] = self.tube_outer_diameter
        var_dict["Tube Inner Diameter"] = self.tube_inner_diameter
        var_dict["Number of Tubes"] = self.number_of_tubes

        return {"vars": var_dict}

    def _get_stream_table_contents(self, time_point=0):
        # Get names for hot and cold sides
        hot_name = self.config.hot_side_name
        cold_name = self.config.cold_side_name
        return create_stream_table_dataframe(
            {
                f"{hot_name} Inlet": self.hot_side_inlet,
                f"{hot_name} Outlet": self.hot_side_outlet,
                f"{cold_name} Inlet": self.cold_side_inlet,
                f"{cold_name} Outlet": self.cold_side_outlet,
            },
            time_point=time_point,
        )

    def calculate_scaling_factors(self):
        super(UnitModelBlockData, self).calculate_scaling_factors()

        for i, c in self.hot_side_heat_transfer_eq.items():
            iscale.constraint_scaling_transform(
                c,
                iscale.get_scaling_factor(
                    self.hot_side.heat[i], default=1, warning=True
                ),
                overwrite=False,
            )

        for i, c in self.cold_side_heat_transfer_eq.items():
            iscale.constraint_scaling_transform(
                c,
                iscale.get_scaling_factor(
                    self.hot_side.heat[i], default=1, warning=True
                ),
                overwrite=False,
            )

        for i, c in self.heat_conservation.items():
            iscale.constraint_scaling_transform(
                c,
                iscale.get_scaling_factor(
                    self.hot_side.heat[i], default=1, warning=True
                ),
                overwrite=False,
            )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________

import pyomo.common.unittest as unittest
import pytest
from idaes.core.surrogate.pysmo.utils import NumpyEvaluator
from pyomo.environ import ConcreteModel, Param, Var, value, sin, atan, atanh
from pyomo.core import ComponentMap

try:
    import numpy as np

    _numpy_available = True
except ImportError:
    _numpy_available = False


@unittest.skipIf(not _numpy_available, "Test requires numpy")
class TestNumpyEvaluator:
    @pytest.mark.unit
    def test_eval_numpy(self):
        m = ConcreteModel()
        m.p = Param([1, 2], mutable=True)
        m.x = Var()

        data = np.array([[0, -1, 2], [0.1, 0.2, 0.3], [4, 5, 6]])
        cMap = ComponentMap()
        cMap[m.p[1]] = data[0]
        cMap[m.p[2]] = data[1]
        cMap[m.x] = data[2]

        npe = NumpyEvaluator(cMap)

        result = npe.walk_expression(sin(m.x))
        assert pytest.approx(result[0], rel=1e-12) == sin(4)
        assert pytest.approx(result[1], rel=1e-12) == sin(5)
        assert pytest.approx(result[2], rel=1e-12) == sin(6)

        result = npe.walk_expression(abs(m.x * m.p[1] - m.p[2]))
        assert pytest.approx(result[0], rel=1e-12) == 0.1
        assert pytest.approx(result[1], rel=1e-12) == -((-1 * 5) - 0.2)
        assert pytest.approx(result[2], rel=1e-12) == (2 * 6 - 0.3)

        result = npe.walk_expression(atan(m.x))
        assert pytest.approx(result[0], rel=1e-12) == atan(4)
        assert pytest.approx(result[1], rel=1e-12) == atan(5)
        assert pytest.approx(result[2], rel=1e-12) == atan(6)

        result = npe.walk_expression(atanh(m.p[2]))
        assert pytest.approx(result[0], rel=1e-12) == atanh(0.1)
        assert pytest.approx(result[1], rel=1e-12) == atanh(0.2)
        assert pytest.approx(result[2], rel=1e-12) == atanh(0.3)

    @pytest.mark.unit
    def test_eval_constant(self):
        m = ConcreteModel()
        m.p = Param([1, 2], mutable=True)
        m.x = Var(initialize=0.25)

        cMap = ComponentMap()
        cMap[m.p[1]] = 2
        cMap[m.p[2]] = 4

        npe = NumpyEvaluator(cMap)

        expr = m.p[1] + m.p[2] + m.x + 0.5
        assert npe.walk_expression(expr) == pytest.approx(6.75, rel=1e-12)

        m.p[1] = 2
        m.p[2] = 4
        assert value(expr) == pytest.approx(6.75, rel=1e-12)


if __name__ == "__main__":
    pytest.main()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# TODO: Missing doc strings
# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

# TODO: Look into protected access issues
# pylint: disable=protected-access

# stdlib
import io
import json
from json import JSONEncoder, JSONDecodeError
import logging
from typing import Dict, Union

# third-party
import numpy as np
import pandas as pd

# package
import pyomo.core as pc
from pyomo.environ import Constraint, sin, cos, log, exp, Set, Param
from pyomo.common.config import ConfigValue, In, Bool, PositiveInt, PositiveFloat
from idaes.core.surrogate.base.surrogate_base import SurrogateTrainer, SurrogateBase
from idaes.core.surrogate.pysmo import (
    polynomial_regression as pr,
    radial_basis_function as rbf,
    kriging as krg,
)
import idaes.logger as idaeslog
from idaes.core.util import to_json


__author__ = "Oluwamayowa Amusat"
# Logging
# -------
_log = idaeslog.getLogger(__name__)

# Global variables
# ----------------
GLOBAL_FUNCS = {"sin": sin, "cos": cos, "log": log, "exp": exp}


class PysmoSurrogateTrainingResult:
    """
    This is an internal helper class for the PysmoTrainer class.

    It stores the results of a trained PySMO model for a single output.

    The class attributes stored are the model metrics, the PySMO model
    and a string of the surrogate model result.

    """

    def __init__(self):
        self.metrics = {}  # Metrics for the fit
        self._model = None  # Surrogate model object (set/get as property)
        self.expression_str = (
            ""  # Pyomo expression, as a string, for additional variables
        )

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, value):
        """

        This is the internal helper function that sets the 'expression_str' attribute by
        calling Pysmo's ``generate_expression()`` function.

        It also sets the model attribute.

        """
        # variable_names = list(value.get_feature_vector().values())
        if hasattr(value, "regression_data_columns"):
            extra_terms_feature_vector = list(
                value.feature_list[i] for i in value.regression_data_columns
            )
        else:
            extra_terms_feature_vector = list(
                value.feature_list[i] for i in value.x_data_columns
            )

        # extra_terms_feature_vector = list(value.feature_list[i] for i in value.regression_data_columns)
        self.expression_str = str(
            value.generate_expression(list(extra_terms_feature_vector))
        )
        self._model = value


class PysmoTrainedSurrogate:
    """
    A surrogate that is trained for one or more outputs. Contains internal (helper) functions.

    This object is used for saving and loading the surrogate model along with its metrics and expression.

    """

    def __init__(self, model_type=""):
        """
        This function adds and displays the results for a trained PySMO surrogate.

        Args:
            model_type: string representing the model type ('poly', 'kriging' or 'rbf')

        Returns:
            self object with additional attributes representing the input models, output labels, number of outputs and model type.
        """
        self._data = {}
        self.model_type = model_type
        self.num_outputs = 0
        self.output_labels = []
        self.input_labels, self.input_bounds = None, None

    def add_result(self, output_name, result):
        """
        This function updates the ``self._data`` and ``self.output_labels`` attributes with the trained model object and output name respectively.

        """
        self._data[output_name] = result
        self.output_labels.append(output_name)
        # self.num_outputs += 1

    def get_result(self, output_name) -> PysmoSurrogateTrainingResult:
        """
        This function returns the ``PysmoSurrogateTrainingResult`` result object for a apecified output variable.

        Args:
            output_name: output variable label of interest (string)

        Returns:
            ``PysmoSurrogateTrainingResult`` result object for output variable.

        """
        return self._data[output_name]

    def display_pysmo_results(self):
        """
        This function returns a string representation of the surrogate expressions for all output variables
        """
        for k in self.output_labels:
            print(k, ":", self._data[k].expression_str)


# Surrogate trainer classes
# -------------------------


class PysmoTrainer(SurrogateTrainer):
    """Base class for Pysmo surrogate trainer classes."""

    # Initialize with configuration for base SurrogateTrainer
    CONFIG = SurrogateTrainer.CONFIG()

    # Subclasses must override this with a specific surrogate model type name
    model_type = "base"

    def __init__(self, **settings):
        super().__init__(**settings)
        self._trained = PysmoTrainedSurrogate(model_type=self.model_type)

    def train_surrogate(self) -> PysmoTrainedSurrogate:
        """
        General workflow method for training a PySMO surrogate.

        Takes the existing data set and executes the PySMO workflow to create
        a PysmoTrainedSurrogate object containing the trained model based on the current configuration arguments.

        The PySMOTrainedSurrogate object is the expected input for the PysmoSurrogate class.

        Args:
            None

        Returns:
            Python object : an instance of PysmoTrainedSurrogate representing the trained surrogate

        """
        self._trained.num_outputs = len(self._output_labels)
        self._trained.input_labels = self._input_labels
        if hasattr(self, "_input_bounds"):
            self._trained.input_bounds = self._input_bounds
        self._training_main_loop()
        return self._trained

    def _create_model(
        self, pysmo_input: pd.DataFrame, output_label: str
    ) -> Union[pr.PolynomialRegression, rbf.RadialBasisFunctions, krg.KrigingModel]:
        """Subclasses must override this and make it return a PySMO model."""
        raise NotImplementedError(
            "Sub-class fail to implement overload ``_create_model`` method."
        )

    def _get_metrics(self, model) -> Dict:
        """Subclasses should override this to return a dict of metrics for the model."""
        return {}

    def _training_main_loop(self):
        for i, output_label in enumerate(self._output_labels):
            # Create input dataframe
            pysmo_input = pd.concat(
                [
                    self._training_dataframe[self._input_labels],
                    self._training_dataframe[[self._output_labels[i]]],
                ],
                axis=1,
            )
            # Create and train model
            model = self._create_model(pysmo_input, output_label)
            model.training()
            # Store results
            result = PysmoSurrogateTrainingResult()
            result.model = model
            result.metrics = self._get_metrics(model)
            self._trained.add_result(output_label, result)
            # Log the status
            _log.info(f"Model for output {output_label} trained successfully")


class PysmoPolyTrainer(PysmoTrainer):
    """
    Standard SurrogateTrainer for PySMO's polynomial models.

    This defines a set of configuration options for PySMO, along with
    methods to train the polynomial model and return basic metrics (R2, RMSE).

    """

    model_type = "poly"

    CONFIG = PysmoTrainer.CONFIG()

    CONFIG.declare(
        "maximum_polynomial_order",
        ConfigValue(
            default=None,
            domain=PositiveInt,
            description="Maximum order of univariate terms. Maximum value is 10.",
        ),
    )

    CONFIG.declare(
        "number_of_crossvalidations",
        ConfigValue(
            default=3, domain=PositiveInt, description="Number of crossvalidations."
        ),
    )

    CONFIG.declare(
        "training_split",
        ConfigValue(
            default=0.8,
            domain=PositiveFloat,
            description="Training-testing data split for PySMO.",
        ),
    )

    CONFIG.declare(
        "solution_method",
        ConfigValue(
            default=None,
            domain=In(["pyomo", "mle", "bfgs"]),
            description="Method for solving regression problem. Must be one of the options ['pyomo', 'mle', 'bfgs']. ",
        ),
    )

    CONFIG.declare(
        "multinomials",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Option for bi-variate pairwise terms in final polynomial",
        ),
    )

    CONFIG.declare(
        "extra_features",
        ConfigValue(
            default=None,
            domain=list,
            description="List of extra features to be considered for regression (if any), e.g. ['x1 / x2']. ",
        ),
    )

    def _create_model(self, pysmo_input, output_label):
        model = pr.PolynomialRegression(
            pysmo_input,
            pysmo_input,
            maximum_polynomial_order=self.config.maximum_polynomial_order,
            training_split=self.config.training_split,
            solution_method=self.config.solution_method,
            multinomials=self.config.multinomials,
            number_of_crossvalidations=self.config.number_of_crossvalidations,
            overwrite=True,
        )
        variable_headers = model.get_feature_vector()
        if self.config.extra_features is not None:
            # create additional terms
            try:
                add_terms = self.config.extra_features
                for j in model.regression_data_columns:
                    add_terms = [
                        add_terms[k].replace(j, "variable_headers['" + str(j) + "']")
                        for k in range(0, len(add_terms))
                    ]
                # pylint: disable=W0123
                model.set_additional_terms(
                    [
                        eval(m, GLOBAL_FUNCS, {"variable_headers": variable_headers})
                        for m in add_terms
                    ]
                )
            except:
                raise ValueError("Additional features could not be constructed.")
        return model

    def _get_metrics(self, model):
        return {"RMSE": model.errors["MSE"] ** 0.5, "R2": model.errors["R2"]}

    def get_confidence_intervals(
        self, model: PysmoTrainedSurrogate, confidence: float = 0.95
    ) -> Dict:
        """
        Compute confidence intervals for the regression patamaters.

        Args:
            model           : A PysmoTrainedSurrogate object
            confidence      : Required confidence interval level, default = 0.95 (95%)

        Returns:
            dict(<dict>)    : Dictionary object containing confidence intervals for all regressed parameters.

                              The dictionary keys are the output variables originally supplied during model training.

                              The dictionary values are dataframes containing four columns:

                                - Regression coeff.  : The regression coefficients for the trained model
                                - Std. errors        : The standard error on the estimated coefficient
                                - Conf. int. lower   : Lower confidence bounds for the estimated regression parameters
                                - Conf. int. upper   : Upper confidence bounds for the estimated regression parameters
        """
        confint_dict = {}
        for i in model.output_labels:
            confint_dict[i] = model._data[i].model.confint_regression(confidence)
        return confint_dict


class PysmoRBFTrainer(PysmoTrainer):
    """

    Standard SurrogateTrainer for PySMO's Radial Basis Function (RBF) models.

    This defines a set of configuration options for PySMO, along with
    methods to train the polynomial model and return basic metrics (R2, RMSE).

    """

    base_model_type = "rbf"
    model_type = "rbf"

    CONFIG = SurrogateTrainer.CONFIG()

    CONFIG.declare(
        "basis_function",
        ConfigValue(
            default=None,
            domain=In(["linear", "cubic", "gaussian", "mq", "imq", "spline"]),
            description="Basis function for RBF.",
        ),
    )

    CONFIG.declare(
        "solution_method",
        ConfigValue(
            default=None,
            domain=In(["pyomo", "algebraic", "bfgs"]),
            description="Method for solving RBF problem. Must be an instance of 'SolutionMethod (Enum)' ",
        ),
    )

    CONFIG.declare(
        "regularization",
        ConfigValue(
            default=None,
            domain=Bool,
            description="Option for regularization - results in a regression rather than interpolation. "
            "Produces more generalizable models. Useful for noisy data.",
        ),
    )

    def __init__(self, **settings):
        super().__init__(**settings)
        self.model_type = f"{self.config.basis_function} {self.base_model_type}"

    def _create_model(self, pysmo_input, output_label):
        model = rbf.RadialBasisFunctions(
            pysmo_input,
            basis_function=self.config.basis_function,
            solution_method=self.config.solution_method,
            regularization=self.config.regularization,
            overwrite=True,
        )
        model.get_feature_vector()
        return model

    def _get_metrics(self, model) -> Dict:
        return {"R2": model.R2, "RMSE": model.rmse}


class PysmoKrigingTrainer(PysmoTrainer):
    """

    Standard SurrogateTrainer for PySMO's kriging models.

    This defines a set of configuration options for PySMO, along with
    methods to train the polynomial model and return mbasic metrics (R2, RMSE).

    """

    model_type = "kriging"

    CONFIG = PysmoTrainer.CONFIG()

    CONFIG.declare(
        "numerical_gradients",
        ConfigValue(
            default=True,
            domain=Bool,
            description="Choice of whether numerical gradients are used in Kriging model training."
            "Determines choice of optimization algorithm: Basinhopping (False) or BFGS (True)."
            "Using the numerical gradient option leads to quicker (but in complex cases possible sub-optimal)"
            " convergence",
        ),
    )

    CONFIG.declare(
        "regularization",
        ConfigValue(
            default=True,
            domain=Bool,
            description="Option for regularization - results in a regression rather than interpolation. "
            "Produces more generalizable models. Useful for noisy data.",
        ),
    )

    def _create_model(self, pysmo_input, output_label):
        model = krg.KrigingModel(
            pysmo_input,
            numerical_gradients=self.config.numerical_gradients,
            regularization=self.config.regularization,
            overwrite=True,
        )
        model.get_feature_vector()
        return model

    def _get_metrics(self, model):
        return {"RMSE": model.training_rmse, "R2": model.training_R2}


class PysmoSurrogate(SurrogateBase):
    """PySMO surrogate model API."""

    def __init__(
        self,
        trained_surrogates: PysmoTrainedSurrogate,
        input_labels,
        output_labels,
        input_bounds=None,
    ):
        """Create PySMO surrogate model.
        Args:
            trained_surrogates: Results of training surrogates.
            input_labels:
            output_labels:
            input_bounds:
        """
        super().__init__(input_labels, output_labels, input_bounds)
        self._trained = trained_surrogates
        self._trained.input_labels, self._trained.input_bounds = (
            input_labels,
            input_bounds,
        )

    def evaluate_surrogate(self, inputs: pd.DataFrame) -> pd.DataFrame:
        """Evaluate the surrogate model at a set of user-provided values.

        Args:
            inputs: The dataframe of input values to be used in the evaluation.
                The dataframe needs to contain a column corresponding to each of the input labels.
                Additional columns are fine, but are not used.

        Returns:
            output: A dataframe of the the output values evaluated at the provided inputs.
                The index of the output dataframe should match the index of the provided inputs.
        """
        inputdata = inputs[self._input_labels].to_numpy()
        outputs = np.zeros(shape=(inputs.shape[0], len(self._output_labels)))

        for i in range(inputdata.shape[0]):
            row_data = inputdata[i, :].reshape(1, len(self._input_labels))
            for j, output_label in enumerate(self._output_labels):
                result = self._trained.get_result(output_label)
                res_vec = result.model.predict_output(row_data)
                outputs[i, j] = res_vec.item()

        return pd.DataFrame(
            data=outputs, index=inputs.index, columns=self._output_labels
        )

    def populate_block(self, block, additional_options=None):
        """Populate a Pyomo Block with surrogate model constraints.

        Args:
            block: Pyomo Block component to be populated with constraints.
            additional_options: None
                No additional options are required for this surrogate object

        Returns:
            None
        """

        output_set = Set(initialize=self._output_labels, ordered=True)

        def pysmo_rule(b, o):
            in_vars = block.input_vars_as_dict()
            out_vars = block.output_vars_as_dict()
            return out_vars[o] == self._trained.get_result(o).model.generate_expression(
                list(in_vars.values())
            )

        block.pysmo_constraint = Constraint(output_set, rule=pysmo_rule)

    def save(self, stream: io.TextIOBase):
        """Save this surrogate to the provided output stream so the model can be used later.

        Args:
           stream: Output stream for serialized surrogate object.

        Returns:
            None
        """
        # All custom serialization is done by the class provided to json.dump
        json.dump(self._trained, stream, cls=TrainedSurrogateEncoder)

    @classmethod
    def load(cls, stream):
        """Create an instance of a surrogate from a stream.

        Args:
            stream: <_io.StringIO>
                This is the python stream containing the data required to load the surrogate.
                This is often, but does not need to be a string of json data.

        Returns:
            An instance of the derived class or None if it failed to load
        """
        stream.seek(0)
        try:
            json_obj = json.load(
                stream, object_pairs_hook=TrainedSurrogateDecoder.decode_pairs
            )
            obj = PysmoSurrogate(
                trained_surrogates=json_obj[TSEBase.MODEL_KEY],
                input_labels=json_obj[TSEBase.INPUT_KEY],
                output_labels=json_obj[TSEBase.OUTPUT_KEY],
                input_bounds=json_obj[TSEBase.BOUNDS_KEY],
            )
        except JSONDecodeError as err:
            obj = None
            _log.error(f"Error decoding surrogate model (stream={stream}): {err}")
        return obj


# Serialization and deserialization classes
# ------------------------------------------


class TSEBase:
    # top-level keys shared by encoder/decoder classes
    MODEL_KEY = "model_encoding"
    INPUT_KEY = "input_labels"
    OUTPUT_KEY = "output_labels"
    BOUNDS_KEY = "input_bounds"
    TYPE_KEY = "surrogate_type"
    MODEL_ATTR_KEY = "attr"
    MODEL_MAP_KEY = "map"
    MODEL_METRICS_KEY = "errors"


class TrainedSurrogateEncoder(JSONEncoder, TSEBase):
    # Attributes to encode
    attrs = {
        "final_polynomial_order",
        "multinomials",
        "optimal_weights_array",
        "extra_terms_feature_vector",
        "additional_term_expressions",
        "regression_data_columns",
        "errors",
        "centres",
        "x_data_columns",
        "x_data_min",
        "x_data_max",
        "basis_function",
        "self.sigma",
        "y_data_min",
        "y_data_max",
        "weights",
        "sigma",
        "regularization_parameter",
        "R2",
        "rmse",
        "optimal_weights",
        "optimal_p",
        "optimal_mean",
        "optimal_variance",
        "optimal_covariance_matrix",
        "covariance_matrix_inverse",
        "optimal_y_mu",
        "training_R2",
        "training_rmse",
        "x_data",
        "x_data_scaled",
    }

    def default(self, obj):
        """
        This function conforms to the interface expected by the json package.
        """
        if isinstance(obj, PysmoTrainedSurrogate):
            return self._encode_surrogate(obj)
        return super().default(obj)

    def _encode_surrogate(self, obj):
        """Encode the trained surrogate for all outputs."""
        models_enc = {
            o: self._encode_model(obj.get_result(o).model) for o in obj.output_labels
        }
        return {
            self.MODEL_KEY: models_enc,
            self.INPUT_KEY: obj.input_labels,
            self.OUTPUT_KEY: obj.output_labels,
            self.BOUNDS_KEY: obj.input_bounds,
            self.TYPE_KEY: obj.model_type,
        }

    def _encode_model(self, model):
        """Encode the surrogate model for a single output."""
        attr_values, attr_types = {}, {}
        # Walk through all attributes in the model that are also in self.attrs
        for name, val in (
            (v, getattr(model, v)) for v in vars(model) if v in self.attrs
        ):
            e = self._encode_attr(val)
            if e is not None:
                attr_values[name], attr_types[name] = self._encode_attr(val)
        return {self.MODEL_ATTR_KEY: attr_values, self.MODEL_MAP_KEY: attr_types}

    @staticmethod
    def _encode_attr(value):
        def is_pyomo_value(v):
            return isinstance(
                v,
                (
                    pc.base.param.ParamData,
                    pc.base.param.Param,
                    pc.expr.numeric_expr.NPV_ProductExpression,
                    pc.expr.numeric_expr.NPV_DivisionExpression,
                ),
            )

        if isinstance(value, np.ndarray):
            return value.tolist(), "numpy"
        elif isinstance(value, (pd.Series, pd.DataFrame)):
            return value.to_json(orient="index"), "pandas"
        elif is_pyomo_value(value):
            return to_json(value, return_dict=True), "pyomo"
        elif isinstance(value, list):
            if len(value) > 0 and is_pyomo_value(value[0]):
                return [str(k) for k in value], "other"  # skip
            return [str(v) for v in value], "list"
        else:
            return value, "str"


class TrainedSurrogateDecoder(TSEBase):
    """Decode a surrogate that was encoded into JSON by :class:`TrainedSurrogateEncoder`."""

    # TODO: Need to test/debug all these methods

    @classmethod
    def decode_pairs(cls, pairs):
        d = {}
        model_json, model_type = None, None
        for key, value in pairs:
            if _log.isEnabledFor(logging.DEBUG):
                _log.debug(f"JSON decode pairs. key={key}, value={value}")
            if key == cls.MODEL_KEY:
                model_json = value
            elif key == cls.TYPE_KEY:
                model_type = value
            elif key == cls.BOUNDS_KEY:
                # convert list values into tuples
                if value is None:
                    d[key] = None
                else:
                    d[key] = {
                        k: tuple(v) for k, v in value.items() if value is not None
                    }
            else:
                d[key] = value
        if model_json is None:
            # If this is not present, assume that we are not at the outer level, so stop here
            return d
        if model_type is None:
            raise JSONDecodeError(
                f"Missing key '{cls.TYPE_KEY}' in JSON object, so type of model is unknown",
                doc="",
                pos=0,
            )
        # choose decoder for model type
        base_type = model_type.split(" ")[-1]  # for "<foo> rbf" types, picks out 'rbf'
        try:
            model_decoder = getattr(cls, f"_decode_{base_type}_model")
        except AttributeError:
            raise JSONDecodeError(
                f"No decoder found for model type '{base_type}'",
                doc="",
                pos=0,
            )
        # decode model
        _log.info(f"Decode surrogate. type={model_type}")
        trained = PysmoTrainedSurrogate(model_type)
        for output_label, model_data in model_json.items():
            surr_mod = model_decoder(
                model_type,
                model_data[cls.MODEL_ATTR_KEY],
                model_data[cls.MODEL_MAP_KEY],
            )
            result = PysmoSurrogateTrainingResult()
            result.model = surr_mod
            # result.metrics = model_data[cls.MODEL_ATTR_KEY][cls.MODEL_METRICS_KEY]
            trained.add_result(output_label, result)
        d[cls.MODEL_KEY] = trained
        # return decoded dict
        return d

    @classmethod
    def _decode_poly_model(cls, type_, attr, mapping):
        # Construct model with minimal arguments necessary
        columns = ["x", "y"]  # attr["regression_data_columns"]
        orig_data = pd.DataFrame(
            {c: list(range(10)) for c in columns}
        )  # cls.pd_decode(attr["original_data"])
        regr_data = pd.DataFrame(
            {c: list(range(10)) for c in columns}
        )  # cls.pd_decode(attr["regression_data"])
        max_order = 1  # int(attr["final_polynomial_order"])
        _log.debug("Reconstructing PolynomialRegression surrogate.")
        model = pr.PolynomialRegression(orig_data, regr_data, max_order)
        # Set model attributes from saved attributes
        _log.debug("Setting attributes on constructed surrogate model")
        for k, v in attr.items():
            if k not in [
                "feature_list",
                "extra_terms_feature_vector",
                "additional_term_expressions",
            ]:
                type_ = mapping[k]
                decoder = decoders.get(type_, null_decode)
                if _log.isEnabledFor(logging.DEBUG):
                    _log.debug(f"Decode attribute. decoder={decoder}, attr={k}:{type_}")
                decoded_value = decoder(v)
                setattr(model, k, decoded_value)
        p = Param(model.regression_data_columns, mutable=True, initialize=0)
        p.index_set().construct()
        p.construct()
        model.feature_list = p
        model.extra_terms_feature_vector = list(
            model.feature_list[i] for i in model.regression_data_columns
        )
        # Re-create function objects from additional terms
        list_terms = cls._poly_decode_vars(attr["additional_term_expressions"], p)
        # pylint: disable=W0123
        model.additional_term_expressions = [
            eval(m, GLOBAL_FUNCS, {"p": p}) for m in list_terms
        ]

        # Remove all attributes created by default
        for jk in list(vars(model).keys()):
            if jk not in attr.keys() and jk != "feature_list":
                delattr(model, jk)

        # Done: return new model
        return model

    @staticmethod
    def _poly_decode_vars(xyz, p):
        """Decode the variables in a polynomial expression."""
        list_idx_vars = [p._data[i].local_name for i in p._data.keys()]
        list_vars = ['p["' + str(i) + '"]' for i in p.keys()]
        pyomo_vars_expr = xyz
        for i in range(0, len(list_idx_vars)):
            pyomo_vars_expr = [
                var_name.replace(list_idx_vars[i], list_vars[i])
                for var_name in pyomo_vars_expr
            ]
        #  return [eval(r, {}, {"p":p}) for r in pyomo_vars_expr]
        return pyomo_vars_expr

    @classmethod
    def _decode_rbf_model(cls, type_, attr, mapping):
        # Construct model with minimal arguments necessary
        columns = ["x", "y"]
        XY_data = pd.DataFrame({c: list(range(10)) for c in columns})
        model = rbf.RadialBasisFunctions(
            XY_data, basis_function=attr["basis_function"]
        )  # XXX
        # Set model attributes from saved attributes
        for k, v in attr.items():
            setattr(model, k, decoders.get(mapping[k], null_decode)(v))
        p = Param(model.x_data_columns, mutable=True, initialize=0)
        p.index_set().construct()
        p.construct()
        model.feature_list = p

        # Remove all attributes created by default
        for jk in list(vars(model).keys()):
            if jk not in attr.keys() and jk != "feature_list":
                delattr(model, jk)
        # Done: return new model
        return model

    @classmethod
    def _decode_kriging_model(cls, type_, attr, mapping):
        # Construct model with minimal arguments necessary
        columns = ["x", "y"]
        XY_data = pd.DataFrame({c: list(range(10)) for c in columns})
        model = krg.KrigingModel(XY_data)
        # Set model attributes from saved attributes
        for k, v in attr.items():
            setattr(model, k, decoders.get(mapping[k], null_decode)(v))
        p = Param(model.x_data_columns, mutable=True, initialize=0)
        p.index_set().construct()
        p.construct()
        model.feature_list = p

        # Remove all attributes created by default
        for jk in list(vars(model).keys()):
            if jk not in attr.keys() and jk != "feature_list":
                delattr(model, jk)

        # Done: return new model
        return model


# Helper methods for decoding Pandas and Numpy


def pd_decode(v):
    return pd.read_json(v, orient="index")


def numpy_decode(v):
    return np.array(v)


def null_decode(v):
    return v


decoders = {"numpy": numpy_decode, "pandas": pd_decode}



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Author: Andrew Lee, Alejandro Garciadiego
"""

import pytest
from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    Constraint,
    Set,
    value,
    Var,
    units as pyunits,
)
from pyomo.util.check_units import assert_units_consistent
from pyomo.common.unittest import assertStructuredAlmostEqual

from idaes.core import (
    MaterialBalanceType,
    EnergyBalanceType,
    MaterialFlowBasis,
    Component,
)
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    fixed_variables_set,
    activated_constraints_set,
)
from idaes.core.solvers import get_solver

from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)

from idaes.models.properties.modular_properties.state_definitions import FTPx

from idaes.models.properties.modular_properties.examples.HC_PR_vap import (
    configuration_vap,
)
from idaes.models.properties.modular_properties.eos.ceos import cubic_roots_available


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver("ipopt_v2")


def _as_quantity(x):
    unit = pyunits.get_units(x)
    if unit is None:
        unit = pyunits.dimensionless
    return value(x) * unit._get_pint_unit()


# Test for configuration dictionaries with parameters from Properties of Gases
# and liquids 4th edition
class TestParamBlock(object):
    @pytest.mark.unit
    def test_build(self):
        model = ConcreteModel()
        model.params = GenericParameterBlock(**configuration_vap)

        assert isinstance(model.params.phase_list, Set)
        assert len(model.params.phase_list) == 1
        for i in model.params.phase_list:
            assert i in ["Vap"]
        assert model.params.Vap.is_vapor_phase()

        assert isinstance(model.params.component_list, Set)
        assert len(model.params.component_list) == 13
        for i in model.params.component_list:
            assert i in [
                "hydrogen",
                "methane",
                "ethane",
                "propane",
                "nbutane",
                "ibutane",
                "ethylene",
                "propene",
                "butene",
                "pentene",
                "hexene",
                "heptene",
                "octene",
            ]

            assert isinstance(model.params.get_component(i), Component)

        assert isinstance(model.params._phase_component_set, Set)
        assert len(model.params._phase_component_set) == 13
        for i in model.params._phase_component_set:
            assert i in [
                ("Vap", "hydrogen"),
                ("Vap", "methane"),
                ("Vap", "ethane"),
                ("Vap", "propane"),
                ("Vap", "nbutane"),
                ("Vap", "ibutane"),
                ("Vap", "ethylene"),
                ("Vap", "propene"),
                ("Vap", "butene"),
                ("Vap", "pentene"),
                ("Vap", "hexene"),
                ("Vap", "heptene"),
                ("Vap", "octene"),
            ]

        assert model.params.config.state_definition == FTPx

        assertStructuredAlmostEqual(
            model.params.config.state_bounds,
            {
                "flow_mol": (0, 100, 1000, pyunits.mol / pyunits.s),
                "temperature": (273.15, 300, 1500, pyunits.K),
                "pressure": (5e4, 1e5, 1e7, pyunits.Pa),
            },
            item_callback=_as_quantity,
        )

        assert model.params.pressure_ref.value == 101325
        assert model.params.temperature_ref.value == 298.15

        assert model.params.hydrogen.mw.value == 2.016e-3
        assert model.params.hydrogen.pressure_crit.value == 12.9e5
        assert model.params.hydrogen.temperature_crit.value == 33.2

        assert model.params.methane.mw.value == 16.043e-3
        assert model.params.methane.pressure_crit.value == 46e5
        assert model.params.methane.temperature_crit.value == 190.4

        assert model.params.ethane.mw.value == 30.070e-3
        assert model.params.ethane.pressure_crit.value == 48.8e5
        assert model.params.ethane.temperature_crit.value == 305.4

        assert model.params.propane.mw.value == 44.094e-3
        assert model.params.propane.pressure_crit.value == 42.5e5
        assert model.params.propane.temperature_crit.value == 369.8

        assert model.params.nbutane.mw.value == 58.124e-3
        assert model.params.nbutane.pressure_crit.value == 38.0e5
        assert model.params.nbutane.temperature_crit.value == 425.2

        assert model.params.ibutane.mw.value == 58.124e-3
        assert model.params.ibutane.pressure_crit.value == 36.5e5
        assert model.params.ibutane.temperature_crit.value == 408.2

        assert model.params.ethylene.mw.value == 28.054e-3
        assert model.params.ethylene.pressure_crit.value == 50.5e5
        assert model.params.ethylene.temperature_crit.value == 282.4

        assert model.params.propene.mw.value == 42.081e-3
        assert model.params.propene.pressure_crit.value == 46.2e5
        assert model.params.propene.temperature_crit.value == 365.0

        assert model.params.butene.mw.value == 56.104e-3
        assert model.params.butene.pressure_crit.value == 40.2e5
        assert model.params.butene.temperature_crit.value == 419.3

        assert model.params.pentene.mw.value == 70.135e-3
        assert model.params.pentene.pressure_crit.value == 40.5e5
        assert model.params.pentene.temperature_crit.value == 464.7

        assert model.params.hexene.mw.value == 84.162e-3
        assert model.params.hexene.pressure_crit.value == 31.7e5
        assert model.params.hexene.temperature_crit.value == 504.0

        assert model.params.heptene.mw.value == 98.189e-3
        assert model.params.heptene.pressure_crit.value == 25.4e5
        assert model.params.heptene.temperature_crit.value == 537.2

        assert model.params.octene.mw.value == 112.216e-3
        assert model.params.octene.pressure_crit.value == 26.2e5
        assert model.params.octene.temperature_crit.value == 566.6
        assert_units_consistent(model)


class TestStateBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.params = GenericParameterBlock(**configuration_vap)

        model.props = model.params.build_state_block([1], defined_state=True)

        # Fix state
        model.props[1].flow_mol.fix(1)
        model.props[1].temperature.fix(295.00)
        model.props[1].pressure.fix(1e5)
        model.props[1].mole_frac_comp["hydrogen"].fix(0.077)
        model.props[1].mole_frac_comp["methane"].fix(0.077)
        model.props[1].mole_frac_comp["ethane"].fix(0.077)
        model.props[1].mole_frac_comp["propane"].fix(0.077)
        model.props[1].mole_frac_comp["nbutane"].fix(0.077)
        model.props[1].mole_frac_comp["ibutane"].fix(0.077)
        model.props[1].mole_frac_comp["ethylene"].fix(0.077)
        model.props[1].mole_frac_comp["propene"].fix(0.077)
        model.props[1].mole_frac_comp["butene"].fix(0.077)
        model.props[1].mole_frac_comp["pentene"].fix(0.077)
        model.props[1].mole_frac_comp["hexene"].fix(0.077)
        model.props[1].mole_frac_comp["heptene"].fix(0.077)
        model.props[1].mole_frac_comp["octene"].fix(0.076)

        assert degrees_of_freedom(model.props[1]) == 0

        return model

    @pytest.mark.unit
    def test_build(self, model):
        # Check state variable values and bounds
        assert isinstance(model.props[1].flow_mol, Var)
        assert value(model.props[1].flow_mol) == 1
        assert model.props[1].flow_mol.ub == 1000
        assert model.props[1].flow_mol.lb == 0

        assert isinstance(model.props[1].pressure, Var)
        assert value(model.props[1].pressure) == 1e5
        assert model.props[1].pressure.ub == 1e7
        assert model.props[1].pressure.lb == 5e4

        assert isinstance(model.props[1].temperature, Var)
        assert value(model.props[1].temperature) == 295
        assert model.props[1].temperature.ub == 1500
        assert model.props[1].temperature.lb == 273.15

        assert isinstance(model.props[1].mole_frac_comp, Var)
        assert len(model.props[1].mole_frac_comp) == 13
        for i in model.props[1].mole_frac_comp:
            assert value(model.props[1].mole_frac_comp[i]) == pytest.approx(
                0.077, abs=1e-2
            )

        # Check supporting variables
        assert isinstance(model.props[1].flow_mol_phase, Var)
        assert len(model.props[1].flow_mol_phase) == 1

        assert isinstance(model.props[1].mole_frac_phase_comp, Var)
        assert len(model.props[1].mole_frac_phase_comp) == 13

        assert isinstance(model.props[1].phase_frac, Var)
        assert len(model.props[1].phase_frac) == 1

        assert isinstance(model.props[1].total_flow_balance, Constraint)
        assert len(model.props[1].total_flow_balance) == 1

        assert isinstance(model.props[1].component_flow_balances, Constraint)
        assert len(model.props[1].component_flow_balances) == 13

        assert isinstance(model.props[1].phase_fraction_constraint, Constraint)
        assert len(model.props[1].phase_fraction_constraint) == 1

        assert_units_consistent(model)

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.unit
    def test_get_material_flow_terms(self, model):
        for p in model.params.phase_list:
            for j in model.params.component_list:
                assert str(model.props[1].get_material_flow_terms(p, j)) == str(
                    model.props[1].flow_mol_phase_comp[p, j]
                )

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.unit
    def test_get_enthalpy_flow_terms(self, model):
        for p in model.params.phase_list:
            assert str(model.props[1].get_enthalpy_flow_terms(p)) == str(
                model.props[1]._enthalpy_flow_term[p]
            )

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.unit
    def test_get_material_density_terms(self, model):
        for p in model.params.phase_list:
            for j in model.params.component_list:
                if j == "hydrogen":
                    assert str(model.props[1].get_material_density_terms(p, j)) == str(
                        model.props[1]._material_density_term[p, j]
                    )

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.unit
    def test_get_energy_density_terms(self, model):
        for p in model.params.phase_list:
            assert str(model.props[1].get_energy_density_terms(p)) == str(
                model.props[1]._energy_density_term[p]
            )

    @pytest.mark.unit
    def test_default_material_balance_type(self, model):
        assert (
            model.props[1].default_material_balance_type()
            == MaterialBalanceType.componentTotal
        )

    @pytest.mark.unit
    def test_default_energy_balance_type(self, model):
        assert (
            model.props[1].default_energy_balance_type()
            == EnergyBalanceType.enthalpyTotal
        )

    @pytest.mark.unit
    def test_get_material_flow_basis(self, model):
        assert model.props[1].get_material_flow_basis() == MaterialFlowBasis.molar

    @pytest.mark.unit
    def test_define_state_vars(self, model):
        sv = model.props[1].define_state_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in ["flow_mol", "mole_frac_comp", "temperature", "pressure"]

    @pytest.mark.unit
    def test_define_port_members(self, model):
        sv = model.props[1].define_state_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in ["flow_mol", "mole_frac_comp", "temperature", "pressure"]

    @pytest.mark.unit
    def test_define_display_vars(self, model):
        sv = model.props[1].define_display_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in [
                "Total Molar Flowrate",
                "Total Mole Fraction",
                "Temperature",
                "Pressure",
            ]

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, model):
        orig_fixed_vars = fixed_variables_set(model)
        orig_act_consts = activated_constraints_set(model)

        model.props.initialize(optarg={"tol": 1e-6})

        assert degrees_of_freedom(model) == 0

        fin_fixed_vars = fixed_variables_set(model)
        fin_act_consts = activated_constraints_set(model)

        assert len(fin_act_consts) == len(orig_act_consts)
        assert len(fin_fixed_vars) == len(orig_fixed_vars)

        for c in fin_act_consts:
            assert c in orig_act_consts
        for v in fin_fixed_vars:
            assert v in orig_fixed_vars

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        # Check phase equilibrium results
        assert model.props[1].mole_frac_phase_comp[
            "Vap", "hydrogen"
        ].value == pytest.approx(0.077, abs=1e-4)
        assert model.props[1].mole_frac_phase_comp[
            "Vap", "propene"
        ].value == pytest.approx(0.077, abs=1e-4)
        assert model.props[1].phase_frac["Vap"].value == pytest.approx(1.000, abs=1e-4)

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, model):
        model.props[1].report()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for rate forms
"""

import pytest
import types

from pyomo.environ import Block, ConcreteModel, Var, units as pyunits
from pyomo.common.config import ConfigBlock
from pyomo.util.check_units import assert_units_equivalent

from idaes.models.properties.modular_properties.reactions.dh_rxn import *

from idaes.core.util.testing import PhysicalParameterTestBlock
from idaes.core.util.misc import add_object_reference
from idaes.core.base.property_meta import PropertyClassMetadata, UnitSet
from idaes.core import MaterialFlowBasis


@pytest.fixture
def model():
    m = ConcreteModel()

    # # Add a test thermo package for validation
    m.pparams = PhysicalParameterTestBlock()
    m.thermo = m.pparams.build_state_block([1])

    # Create a dummy reaction parameter block
    m.rparams = Block()

    m.rparams.config = ConfigBlock(implicit=True)

    m.rparams.config.property_package = m.pparams
    m.rparams.config.reaction_basis = MaterialFlowBasis.molar
    m.rparams.config.rate_reactions = ConfigBlock(implicit=True)
    m.rparams.config.rate_reactions.r1 = ConfigBlock(implicit=True)
    m.rparams.config.rate_reactions.r1 = {
        "stoichiometry": {("p1", "c1"): -1, ("p1", "c2"): 2},
        "parameter_data": {},
    }
    m.rparams.config.equilibrium_reactions = ConfigBlock(implicit=True)
    m.rparams.config.equilibrium_reactions.e1 = ConfigBlock(implicit=True)
    m.rparams.config.equilibrium_reactions.e1 = {
        "stoichiometry": {("p1", "c1"): -1, ("p1", "c2"): 2},
        "parameter_data": {},
    }

    m.rparams.reaction_r1 = Block()
    m.rparams.reaction_e1 = Block()

    m.meta_object = PropertyClassMetadata()
    m.meta_object._default_units.set_units(
        temperature=pyunits.K,
        mass=pyunits.kg,
        length=pyunits.m,
        time=pyunits.s,
        amount=pyunits.mol,
    )

    def get_metadata(self):
        return m.meta_object

    m.rparams.get_metadata = types.MethodType(get_metadata, m.rparams)

    # Create a dummy state block
    m.rxn = Block([1])
    add_object_reference(m.rxn[1], "params", m.rparams)
    add_object_reference(m.rxn[1], "state_ref", m.thermo[1])

    return m


@pytest.mark.unit
def test_constant_dh_rxn(model):
    model.rparams.config.rate_reactions.r1.parameter_data = {"dh_rxn_ref": 1}
    model.rparams.config.equilibrium_reactions.e1.parameter_data = {"dh_rxn_ref": 10}

    constant_dh_rxn.build_parameters(
        model.rparams.reaction_r1, model.rparams.config.rate_reactions["r1"]
    )
    constant_dh_rxn.build_parameters(
        model.rparams.reaction_e1, model.rparams.config.equilibrium_reactions["e1"]
    )

    # Check parameter construction
    assert isinstance(model.rparams.reaction_r1.dh_rxn_ref, Var)
    assert model.rparams.reaction_r1.dh_rxn_ref.value == 1

    assert isinstance(model.rparams.reaction_e1.dh_rxn_ref, Var)
    assert model.rparams.reaction_e1.dh_rxn_ref.value == 10

    # Check expressions
    rform = constant_dh_rxn.return_expression(
        model.rxn[1], model.rparams.reaction_r1, "r1", 300
    )

    assert str(rform) == str(model.rparams.reaction_r1.dh_rxn_ref)

    rform = constant_dh_rxn.return_expression(
        model.rxn[1], model.rparams.reaction_e1, "e1", 300
    )

    assert str(rform) == str(model.rparams.reaction_e1.dh_rxn_ref)

    assert_units_equivalent(rform, pyunits.J / pyunits.mol)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for flowsheet_model.

Author: Andrew Lee
"""
import pytest
import inspect
from pyomo.environ import ConcreteModel, Constraint, Set, Var, units as pyunits
from pyomo.common.config import ConfigBlock
from idaes.core import (
    declare_process_block_class,
    ReactionParameterBlock,
    ReactionBlockBase,
    ReactionBlockDataBase,
    PhysicalParameterBlock,
    StateBlock,
    StateBlockData,
)
from idaes.core.util.exceptions import PropertyPackageError, PropertyNotSupportedError


# -----------------------------------------------------------------------------
# Test ParameterBlock
@declare_process_block_class("PropertyParameterBlock")
class _PropertyParameterBlock(PhysicalParameterBlock):
    def build(self):
        super(_PropertyParameterBlock, self).build()

    @classmethod
    def define_metadata(cls, obj):
        obj.define_custom_properties(
            {
                "prop1": {"method": None, "units": "m"},
                "prop3": {"method": False, "supported": False},
            }
        )
        obj.add_default_units(
            {
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.g,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            }
        )


@declare_process_block_class("ReactionParameterTestBlock")
class _ReactionParameterBlock(ReactionParameterBlock):
    def build(self):
        super(ReactionParameterBlock, self).build()


@pytest.mark.unit
def test_config_block():
    # Test that PhysicalParameterBlock gets module information
    m = ConcreteModel()
    m.r = ReactionParameterTestBlock()

    assert len(m.r.config) == 2
    assert isinstance(m.r.config.default_arguments, ConfigBlock)
    assert len(m.r.config.default_arguments) == 0


@pytest.mark.unit
def test_ReactionParameter_NotImplementedErrors():
    # Test that class methods return NotImplementedError
    m = ConcreteModel()
    m.r = ReactionParameterTestBlock()

    with pytest.raises(NotImplementedError):
        m.r.get_metadata()


@declare_process_block_class("ReactionParameterBlock2")
class _ReactionParameterBlock2(ReactionParameterBlock):
    def build(self):
        super(ReactionParameterBlock, self).build()

    @classmethod
    def get_required_properties(self):
        return ["prop1"]

    @classmethod
    def define_metadata(cls, obj):
        obj.define_custom_properties({"rxn1": {"method": None}})
        obj.add_default_units(
            {
                "time": pyunits.hr,
                "length": pyunits.m,
                "mass": pyunits.g,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            }
        )


@pytest.mark.unit
def test_validate_state_block_invalid_units():
    # Test validation of associated PropertyParameterBlock
    m = ConcreteModel()
    m.p = PropertyParameterBlock()
    m.r = ReactionParameterBlock2(property_package=m.p)

    with pytest.raises(PropertyPackageError):
        m.r._validate_property_parameter_units()


@declare_process_block_class("ReactionParameterBlock3")
class _ReactionParameterBlock3(ReactionParameterBlock):
    def build(self):
        super(ReactionParameterBlock, self).build()

    @classmethod
    def define_metadata(cls, obj):
        obj.define_custom_properties(
            {"rxn1": {"method": None}, "props2": {"required": True}}
        )
        obj.add_default_units(
            {
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.g,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            }
        )


@pytest.mark.unit
def test_validate_state_block_unsupported_prop():
    # Test validation of associated PropertyParameterBlock
    m = ConcreteModel()
    m.p = PropertyParameterBlock()
    m.r = ReactionParameterBlock3(property_package=m.p)

    with pytest.raises(PropertyPackageError):
        m.r._validate_property_parameter_properties()


@declare_process_block_class("ReactionParameterBlock4")
class _ReactionParameterBlock4(ReactionParameterBlock):
    def build(self):
        super(ReactionParameterBlock, self).build()

    @classmethod
    def define_metadata(cls, obj):
        obj.define_custom_properties(
            {"rxn1": {"method": None}, "prop3": {"required": True}}
        )
        obj.add_default_units(
            {
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.g,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            }
        )


@pytest.mark.unit
def test_validate_state_block_unsupported_prop_False():
    # Test validation of associated PropertyParameterBlock
    m = ConcreteModel()
    m.p = PropertyParameterBlock()
    m.r = ReactionParameterBlock4(property_package=m.p)

    with pytest.raises(PropertyPackageError):
        m.r._validate_property_parameter_properties()


@declare_process_block_class("ReactionParameterBlock5")
class _ReactionParameterBlock5(ReactionParameterBlock):
    def build(self):
        super(ReactionParameterBlock, self).build()

    @classmethod
    def define_metadata(cls, obj):
        obj.define_custom_properties(
            {
                "rxn1": {"method": None},
                "prop1": {"required": True},
            }
        )
        obj.add_default_units(
            {
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.g,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            }
        )


@declare_process_block_class("ReactionParameterBlock6")
class _ReactionParameterBlock6(ReactionParameterBlock):
    def build(self):
        super(ReactionParameterBlock, self).build()

    @classmethod
    def define_metadata(cls, obj):
        obj.define_custom_properties(
            {
                "rxn1": {"method": None},
                "prop1": {"required": True},
            }
        )
        obj.add_default_units(
            {
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.g,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            }
        )


@pytest.mark.unit
def test_ReactionParameterBase_build():
    # Test that ReactionParameterBlock builds correctly
    m = ConcreteModel()
    m.p = PropertyParameterBlock()
    m.r = ReactionParameterBlock6(property_package=m.p)
    super(_ReactionParameterBlock6, m.r).build()


# -----------------------------------------------------------------------------
# Test ReactionBlockBase
@declare_process_block_class("ReactionBlock", block_class=ReactionBlockBase)
class ReactionBlockData(ReactionBlockDataBase):
    def build(self):
        super(ReactionBlockDataBase, self).build()


@pytest.mark.unit
def test_ReactionBlockBase_initialize():
    # Test that ReactionBlockBase initialize method raises NotImplementedError
    m = ConcreteModel()
    # Need to index block so that it does not do multiple inheritance
    m.s = Set(initialize=[1, 2])
    m.r = ReactionBlock(m.s)

    with pytest.raises(NotImplementedError):
        m.r.initialize()


@pytest.mark.unit
def test_ReactionBlockBase_report():
    # Test that ReactionBlockBase report method raises NotImplementedError
    m = ConcreteModel()
    # Need to index block so that it does not do multiple inheritance
    m.s = Set(initialize=[1, 2])
    m.r = ReactionBlock(m.s)

    with pytest.raises(NotImplementedError):
        m.r.report()


# -----------------------------------------------------------------------------
# Test ReactionBlockDataBase
@pytest.mark.unit
def test_StateBlock_config():
    # Test that ReactionBlockDataBase config has correct arguments
    m = ConcreteModel()
    m.p = ReactionBlock()

    assert len(m.p.config) == 3
    assert hasattr(m.p.config, "parameters")
    assert hasattr(m.p.config, "state_block")
    assert hasattr(m.p.config, "has_equilibrium")

    m.p.config.has_equilibrium = True
    m.p.config.has_equilibrium = False
    with pytest.raises(ValueError):
        m.p.config.has_equilibrium = "foo"
    with pytest.raises(ValueError):
        m.p.config.has_equilibrium = 10


@declare_process_block_class("DummyStateBlock", block_class=StateBlock)
class StateTestBlockData(StateBlockData):
    def build(self):
        super(StateBlockData, self).build()


@pytest.mark.unit
def test_validate_state_block_fail():
    # Test that ReactionBlockDataBase validate_state_block returns error
    m = ConcreteModel()
    m.p = PropertyParameterBlock()
    m.p2 = PropertyParameterBlock()

    m.pb = DummyStateBlock(parameters=m.p2)

    m.r = ReactionParameterBlock6(property_package=m.p)
    super(_ReactionParameterBlock6, m.r).build()

    m.rb = ReactionBlock(parameters=m.r, state_block=m.pb)

    with pytest.raises(PropertyPackageError):
        m.rb._validate_state_block()


@declare_process_block_class("ReactionBlock2", block_class=ReactionBlockBase)
class ReactionBlockData2(ReactionBlockDataBase):
    def build(self):
        super(ReactionBlockData2, self).build()


@pytest.mark.unit
def test_build():
    # Test that ReactionBlockDataBase builds correctly with good arguments
    m = ConcreteModel()
    m.p = PropertyParameterBlock()

    m.pb = DummyStateBlock(parameters=m.p)

    m.r = ReactionParameterBlock6(property_package=m.p)
    super(_ReactionParameterBlock6, m.r).build()

    m.rb = ReactionBlock2(parameters=m.r, state_block=m.pb)

    assert hasattr(m.rb, "state_ref")
    assert m.rb.params == m.rb.config.parameters


# -----------------------------------------------------------------------------
# Test reaction __getattr__ method
@declare_process_block_class("Parameters")
class _Parameters(ReactionParameterBlock):
    def build(self):
        super(ReactionParameterBlock, self).build()
        frm = inspect.stack()[1]
        self.__package_module = inspect.getmodule(frm[0])

    @classmethod
    def define_metadata(cls, obj):
        obj.define_custom_properties(
            {
                "a": {"method": "a_method"},
                "recursion1": {"method": "_recursion1"},
                "recursion2": {"method": "_recursion2"},
                "not_callable": {"method": "test_obj"},
                "raise_exception": {"method": "_raise_exception"},
                "not_supported": {"supported": False},
                "does_not_create_component": {"method": "_does_not_create_component"},
            }
        )


@declare_process_block_class("Reaction", block_class=ReactionBlockBase)
class _Reaction(ReactionBlockDataBase):
    def build(self):
        super(ReactionBlockDataBase, self).build()

        self.test_obj = 1

    def a_method(self):
        self.a = Var(initialize=1)

    def _recursion1(self):
        self.recursive_cons1 = Constraint(expr=self.recursion2 == 1)

    def _recursion2(self):
        self.recursive_cons2 = Constraint(expr=self.recursion1 == 1)

    def _raise_exception(self):
        # PYLINT-TODO
        # pylint: disable-next=broad-exception-raised
        raise Exception()

    def _does_not_create_component(self):
        pass


@pytest.fixture()
def m():
    m = ConcreteModel()
    m.pb = Parameters()
    m.p = Reaction(parameters=m.pb)

    return m


@pytest.mark.unit
def test_getattr_add_var(m):
    assert isinstance(m.p.a, Var)
    assert m.p.a.value == 1
    assert m.p.is_property_constructed("a") == True


@pytest.mark.unit
def test_lock_attributes(m):
    with pytest.raises(AttributeError):
        with m.p.lock_attribute_creation_context():
            m.p.a.value == 1
    # Make sure it unlocked
    assert m.p.a.value == 1


@pytest.mark.unit
def test_is_property_constructed(m):
    assert m.p.is_property_constructed("a") == False
    assert m.p.a.value == 1
    assert m.p.is_property_constructed("a") == True


@pytest.mark.unit
def test_getattr_protected(m):
    with pytest.raises(PropertyPackageError):
        # Call a protected component that does not exist
        m.p.cons = Constraint(expr=m.p._foo == 1)


@pytest.mark.unit
def test_getattr_recursion(m):
    with pytest.raises(PropertyPackageError):
        # Call a component that triggers a recursive loop of calls
        m.p.cons = Constraint(expr=m.p.recursion1 == 1)


@pytest.mark.unit
def test_getattr_does_not_exist(m):
    with pytest.raises(PropertyNotSupportedError):
        m.p.cons = Constraint(expr=m.p.does_not_exist == 1)


@pytest.mark.unit
def test_getattr_not_callable(m):
    with pytest.raises(PropertyPackageError):
        m.p.cons = Constraint(expr=m.p.not_callable == 1)


@pytest.mark.unit
def test_getattr_not_supported(m):
    with pytest.raises(PropertyNotSupportedError):
        m.p.cons = Constraint(expr=m.p.not_supported == 1)


@pytest.mark.unit
def test_getattr_raise_exception(m):
    with pytest.raises(Exception):
        m.p.cons = Constraint(expr=m.p.raise_exception == 1)


# TODO : Need a test for cases where method does not create property
# @pytest.mark.unit
# def test_getattr_does_not_create_component(m):
#    with pytest.raises(PropertyPackageError):
#        m.p.cons = Constraint(expr=m.p.does_not_create_component == 1)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Heat Exchanger Models.
"""

__author__ = "John Eslick"

from enum import Enum

# Import Pyomo libraries
from pyomo.environ import (
    Block,
    Var,
    Param,
    log,
    Reference,
    PositiveReals,
    ExternalFunction,
    units as pyunits,
    check_optimal_termination,
)
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    UnitModelBlockData,
)

import idaes.logger as idaeslog
from idaes.core.util.functions import functions_lib
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.models.unit_models.heater import (
    _make_heater_config_block,
    _make_heater_control_volume,
)

from idaes.core.util.misc import add_object_reference
from idaes.core.util import scaling as iscale
from idaes.core.solvers import get_solver
from idaes.core.util.exceptions import ConfigurationError, InitializationError
from idaes.core.initialization import SingleControlVolumeUnitInitializer


_log = idaeslog.getLogger(__name__)


class HeatExchangerFlowPattern(Enum):
    """
    Enum of supported flow patterns for heat exchangers.
    """

    countercurrent = 1
    cocurrent = 2
    crossflow = 3


class HX0DInitializer(SingleControlVolumeUnitInitializer):
    """
    Initializer for 0D Heat Exchanger units.

    """

    def initialization_routine(
        self,
        model: Block,
        plugin_initializer_args: dict = None,
        copy_inlet_state: bool = False,
        duty=1000 * pyunits.W,
    ):
        """
        Common initialization routine for 0D Heat Exchangers.

        This routine starts by initializing the hot and cold side properties. Next, the heat
        transfer between the two sides is fixed to an initial guess for the heat duty (provided by the duty
        argument), the associated constraint is deactivated, and the model is then solved. Finally, the heat
        duty is unfixed and the heat transfer constraint reactivated followed by a final solve of the model.

        Args:
            model: Pyomo Block to be initialized
            plugin_initializer_args: dict-of-dicts containing arguments to be passed to plug-in Initializers.
                Keys should be submodel components.
            copy_inlet_state: bool (default=False). Whether to copy inlet state to other states or not
                (0-D control volumes only). Copying will generally be faster, but inlet states may not contain
                all properties required elsewhere.
            duty: initial guess for heat duty to assist with initialization. Can be a Pyomo expression with units.

        Returns:
            Pyomo solver results object
        """
        return super(SingleControlVolumeUnitInitializer, self).initialization_routine(
            model=model,
            plugin_initializer_args=plugin_initializer_args,
            copy_inlet_state=copy_inlet_state,
            duty=duty,
        )

    def initialize_main_model(
        self,
        model: Block,
        copy_inlet_state: bool = False,
        duty=1000 * pyunits.W,
    ):
        """
        Initialization routine for main 0D HX models.

        Args:
            model: Pyomo Block to be initialized.
            copy_inlet_state: bool (default=False). Whether to copy inlet state to other states or not
                (0-D control volumes only). Copying will generally be faster, but inlet states may not contain
                all properties required elsewhere.
            duty: initial guess for heat duty to assist with initialization, default = 1000 W. Can
                be a Pyomo expression with units.

        Returns:
            Pyomo solver results object.

        """
        # Get loggers
        init_log = idaeslog.getInitLogger(
            model.name, self.get_output_level(), tag="unit"
        )
        solve_log = idaeslog.getSolveLogger(
            model.name, self.get_output_level(), tag="unit"
        )

        # Create solver
        solver = self._get_solver()

        self.initialize_control_volume(model.hot_side, copy_inlet_state)
        init_log.info_high("Initialization Step 1a (hot side) Complete.")

        self.initialize_control_volume(model.cold_side, copy_inlet_state)
        init_log.info_high("Initialization Step 1b (cold side) Complete.")
        # ---------------------------------------------------------------------
        # Solve unit without heat transfer equation
        model.heat_transfer_equation.deactivate()

        # Check to see if heat duty is fixed
        # We will assume that if the first point is fixed, it is fixed at all points
        if not model.cold_side.heat[model.flowsheet().time.first()].fixed:
            cs_fixed = False

            model.cold_side.heat.fix(duty)
            for i in model.hot_side.heat:
                model.hot_side.heat[i].set_value(-duty)
        else:
            cs_fixed = True
            for i in model.hot_side.heat:
                model.hot_side.heat[i].set_value(model.cold_side.heat[i])

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solver.solve(model, tee=slc.tee)
        init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(res)))
        if not cs_fixed:
            model.cold_side.heat.unfix()
        model.heat_transfer_equation.activate()
        # ---------------------------------------------------------------------
        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solver.solve(model, tee=slc.tee)
        init_log.info("Initialization Completed, {}".format(idaeslog.condition(res)))

        return res


def _make_heat_exchanger_config(config):
    """
    Declare configuration options for HeatExchangerData block.
    """
    config.declare(
        "hot_side_name",
        ConfigValue(
            default=None,
            domain=str,
            doc="Hot side name, sets control volume and inlet and outlet names",
        ),
    )
    config.declare(
        "cold_side_name",
        ConfigValue(
            default=None,
            domain=str,
            doc="Cold side name, sets control volume and inlet and outlet names",
        ),
    )
    config.declare(
        "hot_side",
        ConfigBlock(
            description="Config block for hot side",
            doc="""A config block used to construct the hot side control volume.
This config can be given by the hot side name instead of hot_side.""",
        ),
    )
    config.declare(
        "cold_side",
        ConfigBlock(
            description="Config block for cold side",
            doc="""A config block used to construct the cold side control volume.
This config can be given by the cold side name instead of cold_side.""",
        ),
    )
    _make_heater_config_block(config.hot_side)
    _make_heater_config_block(config.cold_side)
    config.declare(
        "delta_temperature_callback",
        ConfigValue(
            default=delta_temperature_lmtd_callback,
            description="Callback for for temperature difference calculations",
        ),
    )
    config.declare(
        "flow_pattern",
        ConfigValue(
            default=HeatExchangerFlowPattern.countercurrent,
            domain=In(HeatExchangerFlowPattern),
            description="Heat exchanger flow pattern",
            doc="""Heat exchanger flow pattern,
**default** - HeatExchangerFlowPattern.countercurrent.
**Valid values:** {
**HeatExchangerFlowPattern.countercurrent** - countercurrent flow,
**HeatExchangerFlowPattern.cocurrent** - cocurrent flow,
**HeatExchangerFlowPattern.crossflow** - cross flow, factor times
countercurrent temperature difference.}""",
        ),
    )


def delta_temperature_lmtd_smooth_callback(b):
    r"""
    This is a callback for a temperature difference expression to calculate
    :math:`\Delta T` in the heat exchanger model using log-mean temperature
    difference (LMTD) approximation from Kazi, S., M. Short, A.J. Isafiade,
    L.T. Biegler. "Heat exchanger network synthesus with detailed exchanger
    designs - 2. Hybrid optimization strategy for synthesis of heat exchanger
    networks". AIChE Journal, 2020.

    .. math::

        \alpha = \frac{\Delta T_2}{\Delta T_1}

    .. math::

        \Delta T = \frac{\Delta T_1 \sqrt{(\alpha - 1)^2 + \varepsilon}}{\sqrt{(\log_e{\alpha})^2 + \varepsilon}}

    where :math:`\Delta T_1` is the temperature difference at the hot inlet end,
    :math:`\Delta T_2` is the temperature difference at the hot outlet end, and
    :math:`\varepsilon` is the smoothing parameter.

    The smoothing parameter (``eps_lmtd_smoothing``) is mutable and the
    default is 1e-10.
    """
    b.eps_lmtd_smoothing = Param(mutable=True, initialize=1e-10)
    eps = b.eps_lmtd_smoothing
    dT1 = b.delta_temperature_in
    dT2 = b.delta_temperature_out

    @b.Expression(b.flowsheet().time)
    def delta_temperature(b, t):
        alpha = dT2[t] / dT1[t]
        return (
            dT1[t]
            * ((alpha - 1) ** 2 + eps) ** (1 / 2)
            / (log(alpha) ** 2 + eps) ** (1 / 2)
        )


def delta_temperature_lmtd_callback(b):
    r"""
    This is a callback for a temperature difference expression to calculate
    :math:`\Delta T` in the heat exchanger model using log-mean temperature
    difference (LMTD).  It can be supplied to "delta_temperature_callback"
    HeatExchanger configuration option.  This form is

    .. math::

        \Delta T = \frac{\Delta T_1 - \Delta T_2}{
            \log_e\left(\frac{\Delta T_1}{\Delta T_2}\right)}

    where :math:`\Delta T_1` is the temperature difference at the hot inlet end
    and :math:`\Delta T_2` is the temperature difference at the hot outlet end.
    """
    dT1 = b.delta_temperature_in
    dT2 = b.delta_temperature_out

    @b.Expression(b.flowsheet().time)
    def delta_temperature(b, t):
        return (dT1[t] - dT2[t]) / log(dT1[t] / dT2[t])


def delta_temperature_lmtd2_callback(b):
    r"""
    This is a callback for a temperature difference expression to calculate
    :math:`\Delta T` in the heat exchanger model using log-mean temperature
    difference (LMTD).  It can be supplied to "delta_temperature_callback"
    HeatExchanger configuration option. This form is

    .. math::

        \Delta T = \frac{\Delta T_2 - \Delta T_1}{
            \log_e\left(\frac{\Delta T_2}{\Delta T_1}\right)}

    where :math:`\Delta T_1` is the temperature difference at the hot inlet end
    and :math:`\Delta T_2` is the temperature difference at the hot outlet end.
    """
    dT1 = b.delta_temperature_in
    dT2 = b.delta_temperature_out

    @b.Expression(b.flowsheet().time)
    def delta_temperature(b, t):
        return (dT2[t] - dT1[t]) / log(dT2[t] / dT1[t])


def delta_temperature_lmtd3_callback(b):
    r"""
    This is a callback for a temperature difference expression to calculate
    :math:`\Delta T` in the heat exchanger model using log-mean temperature
    difference (LMTD).  It can be supplied to "delta_temperature_callback"
    HeatExchanger configuration option. This form is

    .. math::

        \Delta T = \frac{\Delta T_1 - \Delta T_2}{
            \log_e\left(\Delta T_2\right) - \log_e\left(\Delta T_1\right)}

    where :math:`\Delta T_1` is the temperature difference at the hot inlet end
    and :math:`\Delta T_2` is the temperature difference at the hot outlet end.
    """
    dT1 = b.delta_temperature_in
    dT2 = b.delta_temperature_out

    @b.Expression(b.flowsheet().time)
    def delta_temperature(b, t):
        return (dT2[t] - dT1[t]) / (log(dT2[t]) - log(dT1[t]))


def delta_temperature_amtd_callback(b):
    r"""
    This is a callback for a temperature difference expression to calculate
    :math:`\Delta T` in the heat exchanger model using arithmetic-mean
    temperature difference (AMTD).  It can be supplied to
    "delta_temperature_callback" HeatExchanger configuration option. This form is

    .. math::

        \Delta T = \frac{\Delta T_1 + \Delta T_2}{2}

    where :math:`\Delta T_1` is the temperature difference at the hot inlet end
    and :math:`\Delta T_2` is the temperature difference at the hot outlet end.
    """
    dT1 = b.delta_temperature_in
    dT2 = b.delta_temperature_out

    @b.Expression(b.flowsheet().time)
    def delta_temperature(b, t):
        return (dT1[t] + dT2[t]) * 0.5


def delta_temperature_underwood_callback(b):
    r"""
    This is a callback for a temperature difference expression to calculate
    :math:`\Delta T` in the heat exchanger model using log-mean temperature
    difference (LMTD) approximation given by Underwood (1970).  It can be
    supplied to "delta_temperature_callback" HeatExchanger configuration option.
    This uses a cube root function that works with negative numbers returning
    the real negative root. This should always evaluate successfully. This form
    is

    .. math::

        \Delta T = \left(\frac{
            \Delta T_1^\frac{1}{3} + \Delta T_2^\frac{1}{3}}{2}\right)^3

    where :math:`\Delta T_1` is the temperature difference at the hot inlet end
    and :math:`\Delta T_2` is the temperature difference at the hot outlet end.
    """
    dT1 = b.delta_temperature_in
    dT2 = b.delta_temperature_out
    temp_units = pyunits.get_units(dT1[dT1.index_set().first()])

    # external function that returns the real root, for the cube root of negative
    # numbers, so it will return without error for positive and negative dT.
    b.cbrt = ExternalFunction(
        library=functions_lib(), function="cbrt", arg_units=[temp_units]
    )

    @b.Expression(b.flowsheet().time)
    def delta_temperature(b, t):
        return ((b.cbrt(dT1[t]) + b.cbrt(dT2[t])) / 2.0) ** 3 * temp_units


def hx_process_config(self):
    """Check for configuration errors and alternate config option names."""
    config = self.config

    if config.cold_side_name in ["hot_side", "cold_side"]:
        raise ConfigurationError(f"cold_side_name cannot be '{config.cold_side_name}'.")
    if config.hot_side_name in ["hot_side", "cold_side"]:
        raise ConfigurationError(f"hot_side_name cannot be '{config.hot_side_name}'.")

    if (
        config.hot_side_name is not None
        and config.cold_side_name is not None
        and config.hot_side_name == config.cold_side_name
    ):
        raise NameError(
            f"HeatExchanger hot and cold side cannot have the same name "
            f"'{config.hot_side_name}'."
        )

    for o in config:
        if not (o in self.CONFIG or o in [config.hot_side_name, config.cold_side_name]):
            raise KeyError("HeatExchanger config option {} not defined".format(o))

    if config.hot_side_name is not None and config.hot_side_name in config:
        config.hot_side.set_value(config[config.hot_side_name])
        # Allow access to hot_side under the hot_side_name, backward
        # compatible with the tube and shell notation
        setattr(config, config.hot_side_name, config.hot_side)
    if config.cold_side_name is not None and config.cold_side_name in config:
        config.cold_side.set_value(config[config.cold_side_name])
        # Allow access to hot_side under the cold_side_name, backward
        # compatible with the tube and shell notation
        setattr(config, config.cold_side_name, config.cold_side)


def add_hx_references(self):
    """
    Method to add common references for hot and cold sides in heat exchangers.

    Args:
        None

    Returns:
        None
    """
    # Add references to the user provided aliases (if applicable).
    # Using add_object_reference keeps these from showing up when you
    # iterate through pyomo components in a model
    if self.config.hot_side_name is not None:
        if not hasattr(self, self.config.hot_side_name):
            add_object_reference(self, self.config.hot_side_name, self.hot_side)
        else:
            raise ValueError(
                f"{self.name} could not assign hot side alias {self.config.hot_side_name} "
                f"as an attribute of that name already exists."
            )
        if not hasattr(self, self.config.hot_side_name + "_inlet"):
            add_object_reference(
                self, self.config.hot_side_name + "_inlet", self.hot_side_inlet
            )
        else:
            raise ValueError(
                f"{self.name} could not assign hot side inlet alias {self.config.hot_side_name}_inlet "
                f"as an attribute of that name already exists."
            )
        if not hasattr(self, self.config.hot_side_name + "_outlet"):
            add_object_reference(
                self, self.config.hot_side_name + "_outlet", self.hot_side_outlet
            )
        else:
            raise ValueError(
                f"{self.name} could not assign hot side outlet alias {self.config.hot_side_name}_outlet "
                f"as an attribute of that name already exists."
            )
    if self.config.cold_side_name is not None:
        if not hasattr(self, self.config.cold_side_name):
            add_object_reference(self, self.config.cold_side_name, self.cold_side)
        else:
            raise ValueError(
                f"{self.name} could not assign cold side alias {self.config.cold_side_name} "
                f"as an attribute of that name already exists."
            )
        if self.config.cold_side_name is not None:
            if not hasattr(self, self.config.cold_side_name + "_inlet"):
                add_object_reference(
                    self, self.config.cold_side_name + "_inlet", self.cold_side_inlet
                )
            else:
                raise ValueError(
                    f"{self.name} could not assign cold side inlet alias {self.config.cold_side_name}_inlet "
                    f"as an attribute of that name already exists."
                )
            if not hasattr(self, self.config.cold_side_name + "_outlet"):
                add_object_reference(
                    self, self.config.cold_side_name + "_outlet", self.cold_side_outlet
                )
            else:
                raise ValueError(
                    f"{self.name} could not assign cold side outlet alias {self.config.cold_side_name}_outlet "
                    f"as an attribute of that name already exists."
                )


@declare_process_block_class("HeatExchanger", doc="Simple 0D heat exchanger model.")
class HeatExchangerData(UnitModelBlockData):
    """
    Simple 0D heat exchange unit.
    Unit model to transfer heat from one material to another.
    """

    default_initializer = HX0DInitializer

    CONFIG = UnitModelBlockData.CONFIG(implicit=True)
    _make_heat_exchanger_config(CONFIG)

    def build(self):
        """
        Building model

        Args:
            None
        Returns:
            None
        """
        ########################################################################
        #  Call UnitModel.build to setup dynamics and configure                #
        ########################################################################
        super().build()
        hx_process_config(self)
        config = self.config

        ########################################################################
        # Add control volumes                                                  #
        ########################################################################
        hot_side = _make_heater_control_volume(
            self,
            "hot_side",
            config.hot_side,
            dynamic=config.dynamic,
            has_holdup=config.has_holdup,
        )
        cold_side = _make_heater_control_volume(
            self,
            "cold_side",
            config.cold_side,
            dynamic=config.dynamic,
            has_holdup=config.has_holdup,
        )

        ########################################################################
        # Add variables                                                        #
        ########################################################################
        # Use hot side units as basis
        s1_metadata = self.hot_side.config.property_package.get_metadata()

        q_units = s1_metadata.get_derived_units("power")
        u_units = s1_metadata.get_derived_units("heat_transfer_coefficient")
        a_units = s1_metadata.get_derived_units("area")
        temp_units = s1_metadata.get_derived_units("temperature")

        u = self.overall_heat_transfer_coefficient = Var(
            self.flowsheet().time,
            domain=PositiveReals,
            initialize=100.0,
            doc="Overall heat transfer coefficient",
            units=u_units,
        )
        a = self.area = Var(
            domain=PositiveReals,
            initialize=1000.0,
            doc="Heat exchange area",
            units=a_units,
        )
        self.delta_temperature_in = Var(
            self.flowsheet().time,
            initialize=10.0,
            doc="Temperature difference at the hot inlet end",
            units=temp_units,
        )
        self.delta_temperature_out = Var(
            self.flowsheet().time,
            initialize=10.1,
            doc="Temperature difference at the hot outlet end",
            units=temp_units,
        )
        if self.config.flow_pattern == HeatExchangerFlowPattern.crossflow:
            self.crossflow_factor = Var(
                self.flowsheet().time,
                initialize=1.0,
                doc="Factor to adjust countercurrent flow heat "
                "transfer calculation for cross flow.",
            )
            f = self.crossflow_factor

        self.heat_duty = Reference(cold_side.heat)
        ########################################################################
        # Add ports                                                            #
        ########################################################################
        self.add_inlet_port(name="hot_side_inlet", block=hot_side, doc="Hot side inlet")
        self.add_inlet_port(
            name="cold_side_inlet",
            block=cold_side,
            doc="Cold side inlet",
        )
        self.add_outlet_port(
            name="hot_side_outlet", block=hot_side, doc="Hot side outlet"
        )
        self.add_outlet_port(
            name="cold_side_outlet",
            block=cold_side,
            doc="Cold side outlet",
        )
        ########################################################################
        # Add aliases                                                          #
        ########################################################################
        add_hx_references(self)

        ########################################################################
        # Add end temperature difference constraints                           #
        ########################################################################

        @self.Constraint(self.flowsheet().time)
        def delta_temperature_in_equation(b, t):
            if b.config.flow_pattern == HeatExchangerFlowPattern.cocurrent:
                return b.delta_temperature_in[t] == hot_side.properties_in[
                    t
                ].temperature - pyunits.convert(
                    cold_side.properties_in[t].temperature, to_units=temp_units
                )
            else:
                return b.delta_temperature_in[t] == hot_side.properties_in[
                    t
                ].temperature - pyunits.convert(
                    cold_side.properties_out[t].temperature, to_units=temp_units
                )

        @self.Constraint(self.flowsheet().time)
        def delta_temperature_out_equation(b, t):
            if b.config.flow_pattern == HeatExchangerFlowPattern.cocurrent:
                return b.delta_temperature_out[t] == hot_side.properties_out[
                    t
                ].temperature - pyunits.convert(
                    cold_side.properties_out[t].temperature, to_units=temp_units
                )
            else:
                return b.delta_temperature_out[t] == hot_side.properties_out[
                    t
                ].temperature - pyunits.convert(
                    cold_side.properties_in[t].temperature, to_units=temp_units
                )

        ########################################################################
        # Add a unit level energy balance                                      #
        ########################################################################
        @self.Constraint(self.flowsheet().time)
        def unit_heat_balance(b, t):
            return 0 == (
                hot_side.heat[t] + pyunits.convert(cold_side.heat[t], to_units=q_units)
            )

        ########################################################################
        # Add delta T calculations using callback function, lots of options,    #
        #   and users can provide their own if needed                          #
        ########################################################################
        config.delta_temperature_callback(self)
        ########################################################################
        # Add Heat transfer equation                                           #
        ########################################################################
        deltaT = self.delta_temperature

        @self.Constraint(self.flowsheet().time)
        def heat_transfer_equation(b, t):
            if self.config.flow_pattern == HeatExchangerFlowPattern.crossflow:
                return pyunits.convert(self.heat_duty[t], to_units=q_units) == (
                    f[t] * u[t] * a * deltaT[t]
                )
            else:
                return pyunits.convert(self.heat_duty[t], to_units=q_units) == (
                    u[t] * a * deltaT[t]
                )

        ########################################################################
        # Add symbols for LaTeX equation rendering                             #
        ########################################################################
        self.overall_heat_transfer_coefficient.latex_symbol = "U"
        self.area.latex_symbol = "A"
        hot_side.heat.latex_symbol = "Q_1"
        cold_side.heat.latex_symbol = "Q_2"
        self.delta_temperature.latex_symbol = "\\Delta T"

    def initialize_build(
        self,
        state_args_1=None,
        state_args_2=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
        duty=None,
    ):
        """
        Heat exchanger initialization method.

        Args:
            state_args_1 : a dict of arguments to be passed to the property
                initialization for the hot side (see documentation of the specific
                property package) (default = {}).
            state_args_2 : a dict of arguments to be passed to the property
                initialization for the cold side (see documentation of the specific
                property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)
            duty : an initial guess for the amount of heat transferred. This
                should be a tuple in the form (value, units), (default
                = (1000 J/s))

        Returns:
            None

        """
        # Set solver options
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        flags1 = self.hot_side.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, state_args=state_args_1
        )

        init_log.info_high("Initialization Step 1a (hot side) Complete.")

        flags2 = self.cold_side.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, state_args=state_args_2
        )

        init_log.info_high("Initialization Step 1b (cold side) Complete.")
        # ---------------------------------------------------------------------
        # Solve unit without heat transfer equation
        self.heat_transfer_equation.deactivate()

        # Get side 1 and side 2 heat units, and convert duty as needed
        s1_units = self.hot_side.heat.get_units()
        s2_units = self.cold_side.heat.get_units()

        # Check to see if heat duty is fixed
        # WE will assume that if the first point is fixed, it is fixed at all points
        if not self.cold_side.heat[self.flowsheet().time.first()].fixed:
            cs_fixed = False
            if duty is None:
                # Assume 1000 J/s and check for unitless properties
                if s1_units is None and s2_units is None:
                    # Backwards compatibility for unitless properties
                    s1_duty = -1000
                    s2_duty = 1000
                else:
                    s1_duty = pyunits.convert_value(
                        -1000, from_units=pyunits.W, to_units=s1_units
                    )
                    s2_duty = pyunits.convert_value(
                        1000, from_units=pyunits.W, to_units=s2_units
                    )
            else:
                # Duty provided with explicit units
                s1_duty = -pyunits.convert_value(
                    duty[0], from_units=duty[1], to_units=s1_units
                )
                s2_duty = pyunits.convert_value(
                    duty[0], from_units=duty[1], to_units=s2_units
                )

            self.cold_side.heat.fix(s2_duty)
            for i in self.hot_side.heat:
                self.hot_side.heat[i].value = s1_duty
        else:
            cs_fixed = True
            for i in self.hot_side.heat:
                self.hot_side.heat[i].set_value(self.cold_side.heat[i])

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(res)))
        if not cs_fixed:
            self.cold_side.heat.unfix()
        self.heat_transfer_equation.activate()
        # ---------------------------------------------------------------------
        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 3 {}.".format(idaeslog.condition(res)))
        # ---------------------------------------------------------------------
        # Release Inlet state
        self.hot_side.release_state(flags1, outlvl=outlvl)
        self.cold_side.release_state(flags2, outlvl=outlvl)

        init_log.info("Initialization Completed, {}".format(idaeslog.condition(res)))

        if not check_optimal_termination(res):
            raise InitializationError(
                f"{self.name} failed to initialize successfully. Please check "
                f"the output logs for more information."
            )

    def _get_performance_contents(self, time_point=0):
        var_dict = {
            "HX Coefficient": self.overall_heat_transfer_coefficient[time_point]
        }
        var_dict["HX Area"] = self.area
        var_dict["Heat Duty"] = self.heat_duty[time_point]
        if self.config.flow_pattern == HeatExchangerFlowPattern.crossflow:
            var_dict["Crossflow Factor"] = self.crossflow_factor[time_point]
        var_dict["Delta T In"] = self.delta_temperature_in[time_point]
        var_dict["Delta T Out"] = self.delta_temperature_out[time_point]

        expr_dict = {}
        expr_dict["Delta T Driving"] = self.delta_temperature[time_point]

        return {"vars": var_dict, "exprs": expr_dict}

    def _get_stream_table_contents(self, time_point=0):
        # Get names for hot and cold sides
        hot_name = self.config.hot_side_name
        if hot_name is None:
            hot_name = "Hot Side"
        cold_name = self.config.cold_side_name
        if cold_name is None:
            cold_name = "Cold Side"
        return create_stream_table_dataframe(
            {
                f"{hot_name} Inlet": self.hot_side_inlet,
                f"{hot_name} Outlet": self.hot_side_outlet,
                f"{cold_name} Inlet": self.cold_side_inlet,
                f"{cold_name} Outlet": self.cold_side_outlet,
            },
            time_point=time_point,
        )

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        # We have a pretty good idea that the delta Ts will be between about
        # 1 and 100 regardless of process of temperature units, so a default
        # should be fine, so don't warn.  Guessing a typical delta t around 10
        # the default scaling factor is set to 0.1
        sf_dT1 = dict(
            zip(
                self.delta_temperature_in.keys(),
                [
                    iscale.get_scaling_factor(v, default=0.1)
                    for v in self.delta_temperature_in.values()
                ],
            )
        )
        sf_dT2 = dict(
            zip(
                self.delta_temperature_out.keys(),
                [
                    iscale.get_scaling_factor(v, default=0.1)
                    for v in self.delta_temperature_out.values()
                ],
            )
        )

        # U depends a lot on the process and units of measure so user should set
        # this one.
        sf_u = dict(
            zip(
                self.overall_heat_transfer_coefficient.keys(),
                [
                    iscale.get_scaling_factor(v, default=1, warning=True)
                    for v in self.overall_heat_transfer_coefficient.values()
                ],
            )
        )

        # Since this depends on the process size this is another scaling factor
        # the user should always set.
        sf_a = iscale.get_scaling_factor(self.area, default=1, warning=True)

        for t, c in self.heat_transfer_equation.items():
            iscale.constraint_scaling_transform(
                c, sf_dT1[t] * sf_u[t] * sf_a, overwrite=False
            )

        for t, c in self.unit_heat_balance.items():
            iscale.constraint_scaling_transform(
                c, sf_dT1[t] * sf_u[t] * sf_a, overwrite=False
            )

        for t, c in self.delta_temperature_in_equation.items():
            iscale.constraint_scaling_transform(c, sf_dT1[t], overwrite=False)

        for t, c in self.delta_temperature_out_equation.items():
            iscale.constraint_scaling_transform(c, sf_dT2[t], overwrite=False)
